# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io
import os
import sys
import re
import time
import logging
import subprocess

from pyomo.common import Executable
from pyomo.common.collections import Bunch
from pyomo.common.dependencies import attempt_import
from pyomo.common.enums import maximize, minimize
from pyomo.common.errors import ApplicationError
from pyomo.common.fileutils import this_file_dir
from pyomo.common.log import is_debug_set
from pyomo.common.tee import capture_output, TeeStream
from pyomo.common.tempfiles import TempfileManager

from pyomo.opt.base import ProblemFormat, ResultsFormat, OptSolver
from pyomo.opt.base.solvers import _extract_version, SolverFactory
from pyomo.opt.results import (
    SolverStatus,
    TerminationCondition,
    SolutionStatus,
    Solution,
)
from pyomo.opt.solver import ILMLicensedSystemCallSolver
from pyomo.core.kernel.block import IBlock
from pyomo.core import ConcreteModel, Var, Objective

from pyomo.solvers.plugins.solvers.gurobi_direct import gurobipy, gurobipy_available
from pyomo.solvers.plugins.solvers.ASL import ASL

logger = logging.getLogger('pyomo.solvers')
GUROBI_RUN = attempt_import('pyomo.solvers.plugins.solvers.GUROBI_RUN')[0]


@SolverFactory.register('gurobi', doc='The GUROBI LP/MIP solver')
class GUROBI(OptSolver):
    """The GUROBI LP/MIP solver"""

    def __new__(cls, *args, **kwds):
        mode = kwds.pop('solver_io', 'lp')
        if mode is None:
            mode = 'lp'
        #
        if mode == 'lp':
            if gurobipy_available:
                return SolverFactory('_gurobi_file', **kwds)
            else:
                return SolverFactory('_gurobi_shell', **kwds)
        if mode == 'mps':
            if gurobipy_available:
                opt = SolverFactory('_gurobi_file', **kwds)
            else:
                opt = SolverFactory('_gurobi_shell', **kwds)
            opt.set_problem_format(ProblemFormat.mps)
            return opt
        if mode in ['python', 'direct']:
            opt = SolverFactory('gurobi_direct', **kwds)
            if opt is None:
                logger.error('Python API for GUROBI is not installed')
                return
            return opt
        if mode == 'persistent':
            opt = SolverFactory('gurobi_persistent', **kwds)
            if opt is None:
                logger.error('Python API for GUROBI is not installed')
                return
            return opt
        #
        if mode == 'os':
            opt = SolverFactory('_ossolver', **kwds)
        elif mode == 'nl':
            opt = SolverFactory('_gurobi_nl', **kwds)
        else:
            logger.error('Unknown IO type: %s' % mode)
            return
        # The Gurobi ASL solver was 'gurobi_ampl' through Gurobi 11,
        # then was renamed to 'gurobi'.  Check 'gurobi' first, then
        # 'gurobi_ampl'.
        for exe_name in ('gurobi', 'gurobi_ampl'):
            exe = Executable(exe_name)
            if (
                exe.available()
                and b'[-AMPL]'
                in subprocess.run(
                    exe.executable,
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    timeout=1,
                ).stdout
            ):
                opt.set_options(f'solver={exe_name}')
                break
        else:
            # Fall back on 'gurobi' (matching the current Gurobi
            # release) if neither appears to be available.
            opt.set_options('solver=gurobi')
        return opt


@SolverFactory.register('_gurobi_nl', doc='NL interface to the Gurobi solver')
class GUROBINL(ASL):
    """NL interface to gurobi_ampl."""

    def license_is_valid(self):
        m = ConcreteModel()
        m.x = Var(bounds=(1, 2))
        m.obj = Objective(expr=m.x)
        try:
            with capture_output():
                self.solve(m)
            return abs(m.x.value - 1) <= 1e-4
        except:
            return False


@SolverFactory.register(
    '_gurobi_shell', doc='Shell interface to the GUROBI LP/MIP solver'
)
class GUROBISHELL(ILMLicensedSystemCallSolver):
    """Shell interface to the GUROBI LP/MIP solver"""

    _solver_info_cache = {}

    def __init__(self, **kwds):
        #
        # Call base class constructor
        #
        kwds['type'] = 'gurobi'
        ILMLicensedSystemCallSolver.__init__(self, **kwds)

        # NOTE: eventually both of the following attributes should be
        # migrated to a common base class.  is the current solve
        # warm-started? a transient data member to communicate state
        # information across the _presolve, _apply_solver, and
        # _postsolve methods.
        self._warm_start_solve = False
        # related to the above, the temporary name of the MST warm-start
        # file (if any).
        self._warm_start_file_name = None

        #
        # Define valid problem formats and associated results formats
        #
        self._valid_problem_formats = [ProblemFormat.cpxlp, ProblemFormat.mps]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.cpxlp] = [ResultsFormat.soln]
        self._valid_result_formats[ProblemFormat.mps] = [ResultsFormat.soln]
        self.set_problem_format(ProblemFormat.cpxlp)

        # Note: Undefined capabilities default to 'None'
        self._capabilities = Bunch()
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def license_is_valid(self):
        """
        Runs a check for a valid Gurobi license using the
        given executable (default is 'gurobi_cl'). All
        output is hidden. If the test fails for any reason
        (including the executable being invalid), then this
        function will return False.
        """
        solver_exec = self.executable()
        if (solver_exec, 'licensed') in self._solver_info_cache:
            return self._solver_info_cache[(solver_exec, 'licensed')]

        if not solver_exec:
            licensed = False
        else:
            executable = os.path.join(os.path.dirname(solver_exec), 'gurobi_cl')
            try:
                rc = subprocess.call(
                    [executable, "--license"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )
            except OSError:
                try:
                    rc = subprocess.run(
                        [solver_exec],
                        input=('import gurobipy; gurobipy.Env().dispose(); quit()'),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        universal_newlines=True,
                    ).returncode
                except OSError:
                    rc = 1
            licensed = not rc

        self._solver_info_cache[(solver_exec, 'licensed')] = licensed
        return licensed

    def _default_results_format(self, prob_format):
        return ResultsFormat.soln

    def warm_start_capable(self):
        return True

    #
    # write a warm-start file in the GUROBI MST format, which is *not*
    # the same as the CPLEX MST format.
    #
    def _warm_start(self, instance):
        from pyomo.core.base import Var

        # for each variable in the symbol_map, add a child to the
        # variables element.  Both continuous and discrete are accepted
        # (and required, depending on other options).
        #
        # **Note**: This assumes that the symbol_map is "clean", i.e.,
        # contains only references to the variables encountered in
        # constraints
        output_index = 0
        if isinstance(instance, IBlock):
            smap = getattr(instance, "._symbol_maps")[self._smap_id]
        else:
            smap = instance.solutions.symbol_map[self._smap_id]
        byObject = smap.byObject
        with open(self._warm_start_file_name, 'w') as mst_file:
            for vdata in instance.component_data_objects(Var):
                if (vdata.value is not None) and (id(vdata) in byObject):
                    name = byObject[id(vdata)]
                    mst_file.write("%s %s\n" % (name, vdata.value))

    # over-ride presolve to extract the warm-start keyword, if specified.
    def _presolve(self, *args, **kwds):
        # create a context in the temporary file manager for
        # this plugin - is "pop"ed in the _postsolve method.
        TempfileManager.push()

        # if the first argument is a string (representing a filename),
        # then we don't have an instance => the solver is being applied
        # to a file.
        self._warm_start_solve = kwds.pop('warmstart', False)
        self._warm_start_file_name = kwds.pop('warmstart_file', None)
        user_warmstart = False
        if self._warm_start_file_name is not None:
            user_warmstart = True

        # the input argument can currently be one of two things: an
        # instance or a filename.  if a filename is provided and a
        # warm-start is indicated, we go ahead and create the temporary
        # file - assuming that the user has already, via some external
        # mechanism, invoked warm_start() with a instance to create the
        # warm start file.
        if self._warm_start_solve and isinstance(args[0], str):
            # we assume the user knows what they are doing...
            pass
        elif self._warm_start_solve and (not isinstance(args[0], str)):
            # assign the name of the warm start file *before* calling
            # the base class presolve - the base class method ends up
            # creating the command line, and the warm start file-name is
            # (obviously) needed there.
            if self._warm_start_file_name is None:
                assert not user_warmstart
                self._warm_start_file_name = TempfileManager.create_tempfile(
                    suffix='.gurobi.mst'
                )

        # let the base class handle any remaining keywords/actions.
        ILMLicensedSystemCallSolver._presolve(self, *args, **kwds)

        # NB: we must let the base class presolve run first so that the
        # symbol_map is actually constructed!

        if (len(args) > 0) and (not isinstance(args[0], str)):
            if len(args) != 1:
                raise ValueError(
                    "GUROBI _presolve method can only handle a single "
                    "problem instance - %s were supplied" % (len(args),)
                )

            # write the warm-start file - currently only supports MIPs.
            # we only know how to deal with a single problem instance.
            if self._warm_start_solve and (not user_warmstart):
                start_time = time.time()
                self._warm_start(args[0])
                end_time = time.time()
                if self._report_timing is True:
                    print(
                        "Warm start write time=%.2f seconds" % (end_time - start_time)
                    )

    def _default_executable(self):
        if sys.platform == 'win32':
            executable = Executable("gurobi.bat")
        else:
            executable = Executable("gurobi.sh")
        if executable:
            return executable.path()
        if gurobipy_available:
            return sys.executable
        logger.warning(
            "Could not locate the 'gurobi' executable, "
            "which is required for solver %s" % self.name
        )
        self.enable = False
        return None

    def _get_version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        solver_exec = self.executable()
        if (solver_exec, 'version') in self._solver_info_cache:
            return self._solver_info_cache[(solver_exec, 'version')]

        if solver_exec is None:
            ver = _extract_version('')
        else:
            results = subprocess.run(
                [solver_exec],
                input=('import gurobipy; print(gurobipy.gurobi.version()); quit()'),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
            ver = None
            try:
                ver = tuple(eval(results.stdout.strip()))
                while len(ver) < 4:
                    ver += (0,)
            except SyntaxError:
                ver = _extract_version('')
        if ver is not None:
            ver = ver[:4]
        self._solver_info_cache[(solver_exec, 'version')] = ver
        return ver

    def create_command_line(self, executable, problem_files):
        #
        # Define log file
        #
        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix='.gurobi.log')

        #
        # Define solution file
        # As indicated above, contains (in XML) both the solution and
        # solver status.
        #
        if self._soln_file is None:
            self._soln_file = TempfileManager.create_tempfile(suffix='.gurobi.txt')

        #
        # Write the GUROBI execution script
        #

        problem_filename = self._problem_files[0]
        solution_filename = self._soln_file
        warmstart_filename = self._warm_start_file_name

        # translate the options into a normal python dictionary, from a
        # pyomo.common.collections.Bunch - the gurobi_run function
        # doesn't know about pyomo, so the translation is necessary
        # (`repr(options)` doesn't produce executable python code)
        options_dict = dict(self.options)

        # NOTE: the gurobi shell is independent of Pyomo python
        #       virtualized environment, so any imports - specifically
        #       that required to get GUROBI_RUN - must be handled
        #       explicitly.
        # NOTE: The gurobi plugin (GUROBI.py) and GUROBI_RUN.py live in
        #       the same directory.
        script = "import sys\n"
        script += "sys.path.append(%r)\n" % (this_file_dir(),)
        script += "import GUROBI_RUN\n"
        script += "soln = GUROBI_RUN.gurobi_run("
        mipgap = float(self.options.mipgap) if self.options.mipgap is not None else None
        for x in (
            problem_filename,
            warmstart_filename,
            None,
            options_dict,
            self._suffixes,
        ):
            script += "%r," % x
        script += ")\n"
        script += "GUROBI_RUN.write_result(soln, %r)\n" % solution_filename
        script += "quit()\n"

        # dump the script and warm-start file names for the
        # user if we're keeping files around.
        if self._keepfiles:
            script_fname = TempfileManager.create_tempfile(suffix='.gurobi.script')
            script_file = open(script_fname, 'w')
            script_file.write(script)
            script_file.close()

            print("Solver script file: '%s'" % script_fname)
            if self._warm_start_solve and (self._warm_start_file_name is not None):
                print("Solver warm-start file: " + self._warm_start_file_name)

        #
        # Define command line
        #
        cmd = [executable]
        if self._timer:
            cmd.insert(0, self._timer)
        return Bunch(cmd=cmd, script=script, log_file=self._log_file, env=None)

    def process_soln_file(self, results):
        # the only suffixes that we extract are constraint duals,
        # constraint slacks, and variable reduced-costs. scan through
        # the solver suffix list and throw an exception if the user has
        # specified any others.
        extract_duals = False
        extract_slacks = False
        extract_rc = False
        for suffix in self._suffixes:
            flag = False
            if re.match(suffix, "dual"):
                extract_duals = True
                flag = True
            if re.match(suffix, "slack"):
                extract_slacks = True
                flag = True
            if re.match(suffix, "rc"):
                extract_rc = True
                flag = True
            if not flag:
                raise RuntimeError(
                    "***The GUROBI solver plugin cannot extract solution suffix="
                    + suffix
                )

        # check for existence of the solution file
        # not sure why we just return - would think that we
        # would want to indicate some sort of error
        if not os.path.exists(self._soln_file):
            return

        soln = Solution()

        # caching for efficiency
        soln_variables = soln.variable
        soln_constraints = soln.constraint

        num_variables_read = 0

        # string compares are too expensive, so simply introduce some
        # section IDs.
        # 0 - unknown
        # 1 - problem
        # 2 - solution
        # 3 - solver

        section = 0  # unknown

        solution_seen = False

        range_duals = {}
        range_slacks = {}

        INPUT = open(self._soln_file, "r")
        for line in INPUT:
            line = line.strip()
            tokens = [token.strip() for token in line.split(":")]
            if tokens[0] == 'section':
                if tokens[1] == 'problem':
                    section = 1
                elif tokens[1] == 'solution':
                    section = 2
                    solution_seen = True
                elif tokens[1] == 'solver':
                    section = 3
            else:
                if section == 2:
                    if tokens[0] == 'var':
                        if tokens[1] != "ONE_VAR_CONSTANT":
                            soln_variables[tokens[1]] = {"Value": float(tokens[2])}
                            num_variables_read += 1
                    elif tokens[0] == 'status':
                        soln.status = getattr(SolutionStatus, tokens[1])
                    elif tokens[0] == 'gap':
                        soln.gap = float(tokens[1])
                    elif tokens[0] == 'objective':
                        if tokens[1].strip() != 'None':
                            soln.objective['__default_objective__'] = {
                                'Value': float(tokens[1])
                            }
                            if results.problem.sense == minimize:
                                results.problem.upper_bound = float(tokens[1])
                            else:
                                results.problem.lower_bound = float(tokens[1])
                    elif tokens[0] == 'constraintdual':
                        name = tokens[1]
                        if name != "c_e_ONE_VAR_CONSTANT":
                            if name.startswith('c_'):
                                soln_constraints.setdefault(tokens[1], {})["Dual"] = (
                                    float(tokens[2])
                                )
                            elif name.startswith('r_l_'):
                                range_duals.setdefault(name[4:], [0, 0])[0] = float(
                                    tokens[2]
                                )
                            elif name.startswith('r_u_'):
                                range_duals.setdefault(name[4:], [0, 0])[1] = float(
                                    tokens[2]
                                )
                    elif tokens[0] == 'constraintslack':
                        name = tokens[1]
                        if name != "c_e_ONE_VAR_CONSTANT":
                            if name.startswith('c_'):
                                soln_constraints.setdefault(tokens[1], {})["Slack"] = (
                                    float(tokens[2])
                                )
                            elif name.startswith('r_l_'):
                                range_slacks.setdefault(name[4:], [0, 0])[0] = float(
                                    tokens[2]
                                )
                            elif name.startswith('r_u_'):
                                range_slacks.setdefault(name[4:], [0, 0])[1] = float(
                                    tokens[2]
                                )
                    elif tokens[0] == 'varrc':
                        if tokens[1] != "ONE_VAR_CONSTANT":
                            soln_variables[tokens[1]]["Rc"] = float(tokens[2])
                    else:
                        setattr(soln, tokens[0], tokens[1])
                elif section == 1:
                    if tokens[0] == 'sense':
                        if tokens[1] == 'minimize':
                            results.problem.sense = minimize
                        elif tokens[1] == 'maximize':
                            results.problem.sense = maximize
                    else:
                        try:
                            val = eval(tokens[1])
                        except:
                            val = tokens[1]
                        setattr(results.problem, tokens[0], val)
                elif section == 3:
                    if tokens[0] == 'status':
                        results.solver.status = getattr(SolverStatus, tokens[1])
                    elif tokens[0] == 'termination_condition':
                        try:
                            results.solver.termination_condition = getattr(
                                TerminationCondition, tokens[1]
                            )
                        except AttributeError:
                            results.solver.termination_condition = (
                                TerminationCondition.unknown
                            )
                    else:
                        setattr(results.solver, tokens[0], tokens[1])

        INPUT.close()

        # For the range constraints, supply only the dual with the largest
        # magnitude (at least one should always be numerically zero)
        for key, (ld, ud) in range_duals.items():
            if abs(ld) > abs(ud):
                soln_constraints['r_l_' + key] = {"Dual": ld}
            else:
                # Use the same key
                soln_constraints['r_l_' + key] = {"Dual": ud}
        # slacks
        for key, (ls, us) in range_slacks.items():
            if abs(ls) > abs(us):
                soln_constraints.setdefault('r_l_' + key, {})["Slack"] = ls
            else:
                # Use the same key
                soln_constraints.setdefault('r_l_' + key, {})["Slack"] = us

        if solution_seen:
            results.solution.insert(soln)

    def _postsolve(self):
        # take care of the annoying GUROBI log file in the current
        # directory.  this approach doesn't seem overly efficient, but
        # python os module functions doesn't accept regular expression
        # directly.
        filename_list = os.listdir(".")
        for filename in filename_list:
            # IMPT: trap the possible exception raised by the file not
            #       existing.  this can occur in pyro environments where
            #       > 1 workers are running GUROBI, and were started
            #       from the same directory.  these logs don't matter
            #       anyway (we redirect everything), and are largely an
            #       annoyance.
            try:
                if re.match(r'gurobi\.log', filename) != None:
                    os.remove(filename)
            except OSError:
                pass

        # let the base class deal with returning results.
        results = ILMLicensedSystemCallSolver._postsolve(self)

        # finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin. does not
        # include, for example, the execution script. but does include
        # the warm-start file.
        TempfileManager.pop(remove=not self._keepfiles)

        return results


@SolverFactory.register(
    '_gurobi_file', doc='LP/MPS file-based direct interface to the GUROBI LP/MIP solver'
)
class GUROBIFILE(GUROBISHELL):
    """Direct LP/MPS file-based interface to the GUROBI LP/MIP solver"""

    def available(self, exception_flag=False):
        if not gurobipy_available:  # this triggers the deferred import
            if exception_flag:
                raise ApplicationError("gurobipy module not importable")
            return False
        if getattr(self, '_available', None) is None:
            self._check_license()
        ans = self._available[0]
        if exception_flag and not ans:
            raise ApplicationError(msg % self.name)
        return ans

    def license_is_valid(self):
        return self.available(False) and self._available[1]

    def _check_license(self):
        licensed = False
        try:
            # Gurobipy writes out license file information when creating
            # the environment
            with capture_output(capture_fd=True):
                m = gurobipy.Model()
            licensed = True
        except gurobipy.GurobiError:
            licensed = False

        self._available = (True, licensed)

    def _get_version(self):
        return (
            gurobipy.GRB.VERSION_MAJOR,
            gurobipy.GRB.VERSION_MINOR,
            gurobipy.GRB.VERSION_TECHNICAL,
        )

    def _default_executable(self):
        # Bogus, but not None (because the test infrastructure disables
        # solvers where the executable() is None)
        return ""

    def create_command_line(self, executable, problem_files):
        #
        # Define log file
        #
        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix='.gurobi.log')

        #
        # Define command line
        #
        return Bunch(cmd=[], script="", log_file=self._log_file, env=None)

    def _apply_solver(self):
        #
        # Execute the command
        #
        if is_debug_set(logger):
            logger.debug("Running %s", self._command.cmd)

        problem_filename = self._problem_files[0]
        warmstart_filename = self._warm_start_file_name

        # translate the options into a normal python dictionary, from a
        # pyutilib SectionWrapper - because the gurobi_run function was
        # originally designed to run in the Python environment
        # distributed in the Gurobi installation (which doesn't know
        # about pyomo) the translation is necessary.
        options_dict = {}
        for key in self.options:
            options_dict[key] = self.options[key]

        # display the log/solver file names prior to execution. this is useful
        # in case something crashes unexpectedly, which is not without precedent.
        if self._keepfiles:
            if self._log_file is not None:
                print("Solver log file: '%s'" % self._log_file)
            if self._problem_files != []:
                print("Solver problem files: %s" % str(self._problem_files))

        sys.stdout.flush()
        ostreams = [io.StringIO()]
        if self._tee:
            ostreams.append(sys.stdout)
        with capture_output(output=TeeStream(*ostreams), capture_fd=False):
            self._soln = GUROBI_RUN.gurobi_run(
                problem_filename, warmstart_filename, None, options_dict, self._suffixes
            )
        self._log = ostreams[0].getvalue()
        self._rc = 0
        sys.stdout.flush()
        return Bunch(rc=self._rc, log=self._log)

    def process_soln_file(self, results):
        # the only suffixes that we extract are constraint duals,
        # constraint slacks, and variable reduced-costs. Scan through
        # the solver suffix list and throw an exception if the user has
        # specified any others.
        extract_duals = False
        extract_slacks = False
        extract_rc = False
        for suffix in self._suffixes:
            flag = False
            if re.match(suffix, "dual"):
                extract_duals = True
                flag = True
            if re.match(suffix, "slack"):
                extract_slacks = True
                flag = True
            if re.match(suffix, "rc"):
                extract_rc = True
                flag = True
            if not flag:
                raise RuntimeError(
                    "***The GUROBI solver plugin cannot extract solution suffix="
                    + suffix
                )

        soln = Solution()

        # caching for efficiency
        soln_variables = soln.variable
        soln_constraints = soln.constraint

        num_variables_read = 0

        # string compares are too expensive, so simply introduce some
        # section IDs.
        # 0 - unknown
        # 1 - problem
        # 2 - solution
        # 3 - solver

        section = 0  # unknown

        solution_seen = False

        range_duals = {}
        range_slacks = {}

        # Copy over the problem info
        for key, val in self._soln['problem'].items():
            setattr(results.problem, key, val)
        if results.problem.sense == 'minimize':
            results.problem.sense = minimize
        elif results.problem.sense == 'maximize':
            results.problem.sense = maximize

        # Copy over the solver info
        for key, val in self._soln['solver'].items():
            setattr(results.solver, key, val)
        results.solver.status = getattr(SolverStatus, results.solver.status)
        try:
            results.solver.termination_condition = getattr(
                TerminationCondition, results.solver.termination_condition
            )
        except AttributeError:
            results.solver.termination_condition = TerminationCondition.unknown

        # Copy over the solution information
        sol = self._soln.get('solution', None)
        if sol:
            if 'status' in sol:
                soln.status = sol['status']
            if 'gap' in sol:
                soln.gap = sol['gap']
            obj = sol.get('objective', None)
            if obj is not None:
                soln.objective['__default_objective__'] = {'Value': obj}
                if results.problem.sense == minimize:
                    results.problem.upper_bound = obj
                else:
                    results.problem.lower_bound = obj
            for name, val in sol.get('var', {}).items():
                if name == "ONE_VAR_CONSTANT":
                    continue
                soln_variables[name] = {"Value": val}
                num_variables_read += 1
            for name, val in sol.get('varrc', {}).items():
                if name == "ONE_VAR_CONSTANT":
                    continue
                soln_variables[name]["Rc"] = val
            for name, val in sol.get('constraintdual', {}).items():
                if name == "c_e_ONE_VAR_CONSTANT":
                    continue
                if name.startswith('c_'):
                    soln_constraints.setdefault(name, {})["Dual"] = val
                elif name.startswith('r_l_'):
                    range_duals.setdefault(name[4:], [0, 0])[0] = val
                elif name.startswith('r_u_'):
                    range_duals.setdefault(name[4:], [0, 0])[1] = val
            for name, val in sol.get('constraintslack', {}).items():
                if name == "c_e_ONE_VAR_CONSTANT":
                    continue
                if name.startswith('c_'):
                    soln_constraints.setdefault(name, {})["Slack"] = val
                elif name.startswith('r_l_'):
                    range_slacks.setdefault(name[4:], [0, 0])[0] = val
                elif name.startswith('r_u_'):
                    range_slacks.setdefault(name[4:], [0, 0])[1] = val

            results.solution.insert(soln)

        # For the range constraints, supply only the dual with the largest
        # magnitude (at least one should always be numerically zero)
        for key, (ld, ud) in range_duals.items():
            if abs(ld) > abs(ud):
                soln_constraints['r_l_' + key] = {"Dual": ld}
            else:
                # Use the same key
                soln_constraints['r_l_' + key] = {"Dual": ud}
        # slacks
        for key, (ls, us) in range_slacks.items():
            if abs(ls) > abs(us):
                soln_constraints.setdefault('r_l_' + key, {})["Slack"] = ls
            else:
                # Use the same key
                soln_constraints.setdefault('r_l_' + key, {})["Slack"] = us



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.appsi.examples import getting_started
import pyomo.common.unittest as unittest
from pyomo.contrib.appsi.cmodel import cmodel_available
from pyomo.contrib import appsi


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestExamples(unittest.TestCase):
    def test_getting_started(self):
        try:
            import numpy as np
        except:
            raise unittest.SkipTest('numpy is not available')
        opt = appsi.solvers.Cplex()
        if not opt.available():
            raise unittest.SkipTest('cplex is not available')
        getting_started.main(plot=False, n_points=10)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr import ExpressionBase, as_numeric
from pyomo.core import Constraint, Objective, TransformationFactory
from pyomo.core.base.var import Var, VarData
from pyomo.core.util import sequence
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation


@TransformationFactory.register(
    'core.remove_fixed_vars',
    doc="Create an equivalent model that omits all fixed variables.",
)
class EliminateFixedVars(IsomorphicTransformation):
    """
    Create an equivalent model that omits all fixed variables.
    """

    def __init__(self, **kwds):
        kwds['name'] = "eliminate_fixed_vars"
        super(EliminateFixedVars, self).__init__(**kwds)

    def _create_using(self, model, **kwds):
        #
        # Clone the model
        #
        M = model.clone()
        #
        # Iterate over the expressions in all objectives and constraints, replacing fixed
        # variables with their associated constants.
        #
        for ctype in [Objective, Constraint]:
            for obj in M.component_map(Objective).values():
                for name in obj:
                    if not obj[name].expr is None:
                        obj[name].expr = self._fix_vars(obj[name].expr, model)
        #
        # Iterate over variables, omitting those that have fixed values
        #
        ctr = 0
        for i in sequence(M.nvariables()):
            var = M.variable(i)
            del M._var[i - 1]
            if var.fixed:
                if var.is_binary():
                    M.statistics.number_of_binary_variables -= 1
                elif var.is_integer():
                    M.statistics.number_of_integer_variables -= 1
                elif var.is_continuous():
                    M.statistics.number_of_continuous_variables -= 1
                M.statistics.number_of_variables -= 1
                del M._label_var_map[var.label]
                del var.component()._data[var.index]
            else:
                M._var[ctr] = var
                var._old_id = var.id
                var.id = ctr
                ctr += 1
        return M

    def _fix_vars(self, expr, model):
        """Walk through the S-expression, fixing variables."""
        # TODO - Change this to use a visitor pattern!
        if expr._args is None:
            return expr
        _args = []
        for i in range(len(expr._args)):
            if isinstance(expr._args[i], ExpressionBase):
                _args.append(self._fix_vars(expr._args[i], model))
            elif (
                isinstance(expr._args[i], Var) or isinstance(expr._args[i], VarData)
            ) and expr._args[i].fixed:
                if expr._args[i].value != 0.0:
                    _args.append(as_numeric(expr._args[i].value))
            else:
                _args.append(expr._args[i])
        expr._args = _args
        return expr



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for pyomo.opt.results
#
import copy
import json
import pickle
import os

from io import StringIO
from os.path import join
from filecmp import cmp

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import yaml, yaml_available
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

import pyomo.opt.results.container as container
from pyomo.opt import SolverResults

currdir = this_file_dir()

ref_yaml_soln = """\
# ==========================================================
# = Solver Results                                         =
# ==========================================================

# ----------------------------------------------------------
#   Problem Information
# ----------------------------------------------------------
Problem: 
- Lower bound: -Infinity
  Upper bound: Infinity
  Number of objectives: 1
  Number of constraints: 24
  Number of variables: 32
  Sense: unknown

# ----------------------------------------------------------
#   Solver Information
# ----------------------------------------------------------
Solver: 
- Status: ok
  Message: PICO Solver\\x3a final f = 88.200000
  Termination condition: optimal
  Id: 0

# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution: 
- number of solutions: 1
  number of solutions displayed: 1
- Status: optimal
  Variable: 
    v4:
      Value: 46.6666666667
    v11:
      Value: 933.333333333
    v12:
      Value: 10000
    v13:
      Value: 10000
    v14:
      Value: 10000
    v15:
      Value: 10000
    v17:
      Value: 100
    v19:
      Value: 100
    v21:
      Value: 100
    v23:
      Value: 100
    v24:
      Value: 46.6666666667
    v25:
      Value: 53.3333333333
    v27:
      Value: 100
    v29:
      Value: 100
    v31:
      Value: 100
  Constraint: 
    c2: 
      Dual: 0.126
  Status description: OPTIMAL SOLUTION FOUND!
"""

ref_json_soln = """\
{
    "Problem": [
        {
            "Lower bound": -Infinity, 
            "Number of constraints": 24, 
            "Number of objectives": 1, 
            "Number of variables": 32, 
            "Sense": "unknown", 
            "Upper bound": Infinity
        }
    ], 
    "Solution": [
        {
            "number of solutions": 1, 
            "number of solutions displayed": 1
        }, 
        {
            "Constraint": {
                "c2": {
                    "Dual": 0.12599999999999997
                }
            }, 
            "Message": "PICO Solver\\\\x3a final f = 88.200000", 
            "Status": "optimal", 
            "Status description": "OPTIMAL SOLUTION FOUND!", 
            "Variable": {
                "v11": {
                    "Value": 933.3333333333336
                }, 
                "v12": {
                    "Value": 10000.0
                }, 
                "v13": {
                    "Value": 10000.0
                }, 
                "v14": {
                    "Value": 10000.0
                }, 
                "v15": {
                    "Value": 10000.0
                }, 
                "v17": {
                    "Value": 100.0
                }, 
                "v19": {
                    "Value": 100.0
                }, 
                "v21": {
                    "Value": 100.0
                }, 
                "v23": {
                    "Value": 100.0
                }, 
                "v24": {
                    "Value": 46.666666666666664
                }, 
                "v25": {
                    "Value": 53.333333333333336
                }, 
                "v27": {
                    "Value": 100.0
                }, 
                "v29": {
                    "Value": 100.0
                }, 
                "v31": {
                    "Value": 100.0
                }, 
                "v4": {
                    "Value": 46.666666666666664
                }
            }
        }
    ], 
    "Solver": [
        {
            "Id": 0, 
            "Message": "PICO Solver\\\\x3a final f = 88.200000", 
            "Status": "ok", 
            "Termination condition": "optimal"
        }
    ]
}
"""


class Test(unittest.TestCase):

    def test_write_solution(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        OUT = StringIO()
        results.write(ostream=OUT)
        self.assertEqual(
            """
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution: 
- number of solutions: 1
  number of solutions displayed: 1
- Status: unknown
  Objective: No values
  Variable: No nonzero values
  Constraint: No values
            """.strip(),
            OUT.getvalue().strip(),
        )

    def test_write_empty_solution(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        OUT = StringIO()
        results.write(num=None, ostream=OUT)
        self.assertEqual(
            """
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution: 
- number of solutions: 1
  number of solutions displayed: 1
- Status: unknown
  Objective: No values
  Variable: No nonzero values
  Constraint: No values
            """.strip(),
            OUT.getvalue().strip(),
        )

    @unittest.skipIf(not yaml_available, "Test requires 'yaml'")
    def test_read_write_yaml_solution(self):
        """Read a SolverResults Object"""
        IN = StringIO(ref_yaml_soln)
        results = SolverResults()
        results.read(istream=IN, format='yaml')
        OUT = StringIO()
        results.write(ostream=OUT, format='yaml')

        self.assertStructuredAlmostEqual(
            yaml.full_load(StringIO(ref_yaml_soln)),
            yaml.full_load(StringIO(OUT.getvalue())),
            allow_second_superset=True,
        )

    @unittest.skipIf(not yaml_available, "Test requires 'yaml'")
    def test_pickle_yaml_solution(self):
        IN = StringIO(ref_yaml_soln)
        results = SolverResults()
        results.read(istream=IN, format='yaml')

        new_results = pickle.loads(pickle.dumps(results))
        OUT = StringIO()
        new_results.write(ostream=OUT, format='yaml')

        self.assertStructuredAlmostEqual(
            yaml.full_load(StringIO(ref_yaml_soln)),
            yaml.full_load(StringIO(OUT.getvalue())),
            allow_second_superset=True,
        )

    @unittest.skipIf(not yaml_available, "Test requires 'yaml'")
    def test_write_read_yaml_file(self):
        IN = StringIO(ref_yaml_soln)
        results = SolverResults()
        results.read(istream=IN, format='yaml')

        with TempfileManager.new_context() as temp:
            fname = temp.create_tempfile('yaml')
            results.write(filename=fname)
            new_results = SolverResults()
            new_results.read(filename=fname)

        self.assertStructuredAlmostEqual(
            results, new_results, allow_second_superset=True
        )

    def test_read_write_json_solution(self):
        IN = StringIO(ref_json_soln)
        results = SolverResults()
        results.read(istream=IN, format='json')
        OUT = StringIO()
        results.write(ostream=OUT, format='json')

        self.assertStructuredAlmostEqual(
            json.loads(ref_json_soln),
            json.loads(OUT.getvalue()),
            allow_second_superset=True,
        )

    def test_pickle_json_solution(self):
        IN = StringIO(ref_json_soln)
        results = SolverResults()
        results.read(istream=IN, format='json')

        new_results = pickle.loads(pickle.dumps(results))
        OUT = StringIO()
        new_results.write(ostream=OUT, format='json')

        self.assertStructuredAlmostEqual(
            json.loads(ref_json_soln),
            json.loads(OUT.getvalue()),
            allow_second_superset=True,
        )

    def test_write_read_json_file(self):
        IN = StringIO(ref_json_soln)
        results = SolverResults()
        results.read(istream=IN, format='json')

        with TempfileManager.new_context() as temp:
            fname = temp.create_tempfile('.jsn')
            results.write(filename=fname)
            new_results = SolverResults()
            new_results.read(filename=fname)

        self.assertStructuredAlmostEqual(
            results, new_results, allow_second_superset=True
        )

    def test_deepcopy_solution(self):
        IN = StringIO(ref_json_soln)
        result = SolverResults()
        result.read(istream=IN, format='json')
        result_copy = copy.deepcopy(result)
        self.assertEqual(list(result.keys()), list(result_copy.keys()))
        self.assertEqual(str(result), str(result_copy))
        self.assertEqual(result, result)
        self.assertEqual(result, result_copy)

    #
    # deleting is not supported right now
    #
    def test_delete_solution(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        results.solution.delete(0)

        OUT = StringIO()
        results.write(ostream=OUT)
        self.assertEqual(
            """
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution: 
- number of solutions: 0
  number of solutions displayed: 0
            """.strip(),
            OUT.getvalue().strip(),
        )

    def test_get_solution(self):
        """Get a solution from a SolverResults object"""
        results = SolverResults()
        soln = results.solution.add()
        self.assertIs(results.solution[0], soln)

    def test_get_solution_attr_error(self):
        """Create an error with a solution suffix"""
        results = SolverResults()
        soln = results.solution.add()
        with self.assertRaisesRegex(
            AttributeError,
            r"Unknown attribute `bad' for object with type <.*\.Solution'>",
        ):
            tmp = soln.bad

    #
    # This is currently allowed, although soln.variable = True is equivalent to
    #   soln.variable.value = True
    #
    def Xtest_set_solution_attr_error(self):
        """Create an error with a solution suffix"""
        try:
            self.soln.variable = True
            self.fail("Expected attribute error failure for 'variable'")
        except AttributeError:
            pass

    def test_soln_write_zeros(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        soln.variable[1]["Value"] = 0.0
        soln.variable[2]["Value"] = 0.0
        soln.variable[4]["Value"] = 0.0

        OUT = StringIO()
        results.write(ostream=OUT)
        self.assertEqual(
            """
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution: 
- number of solutions: 1
  number of solutions displayed: 1
- Status: unknown
  Objective: No values
  Variable: No nonzero values
  Constraint: No values
            """.strip(),
            OUT.getvalue().strip(),
        )

    def test_soln_write_nonzeros(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        soln.variable[1]["Value"] = 1.0
        soln.variable[2]["Value"] = 3.0
        soln.variable[4]["Value"] = 5.0

        OUT = StringIO()
        results.write(ostream=OUT)
        self.assertEqual(
            """
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution: 
- number of solutions: 1
  number of solutions displayed: 1
- Status: unknown
  Objective: No values
  Variable:
    1:
      Value: 1
    2:
      Value: 3
    4:
      Value: 5
  Constraint: No values
            """.strip(),
            OUT.getvalue().strip(),
        )

    def test_soln_suffix_getattr(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        soln.variable[1]["Value"] = 0.0
        soln.variable[2]["Value"] = 0.1
        soln.variable[4]["Value"] = 0.3
        self.assertEqual(soln.variable[4]["Value"], 0.3)
        self.assertEqual(soln.variable[2]["Value"], 0.1)

    def test_soln_suffix_setattr_getattr(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        soln.variable[1]["Value"] = 0.0
        soln.variable[4]["Value"] = 0.3
        soln.variable[4]["Slack"] = 0.4
        self.assertEqual(list(soln.variable.keys()), [1, 2, 4])
        self.assertEqual(soln.variable[1]["Value"], 0.0)
        self.assertEqual(soln.variable[4]["Value"], 0.3)
        self.assertEqual(soln.variable[4]["Slack"], 0.4)

    def test_soln_to_str(self):
        results = SolverResults()
        soln = results.solution.add()
        soln.variable[1] = {"Value": 0}
        soln.variable[2] = {"Value": 0}
        soln.variable[4] = {"Value": 0}

        self.assertEqual(
            """
Solution: 
- number of solutions: 1
  number of solutions displayed: 1
- Status: unknown
  Objective: No values
  Variable: No nonzero values
  Constraint: No values
            """.strip(),
            str(results).strip(),
        )

        soln.variable[1]["Value"] = 1.0
        soln.variable[2]["Value"] = 3.0
        soln.variable[4]["Value"] = 5.0

        self.assertEqual(
            """
Solution: 
- number of solutions: 1
  number of solutions displayed: 1
- Status: unknown
  Objective: No values
  Variable:
    1:
      Value: 1
    2:
      Value: 3
    4:
      Value: 5
  Constraint: No values
            """.strip(),
            str(results).strip(),
        )


class TestContainer(unittest.TestCase):
    def test_declare_and_str(self):
        class LocalContainer(container.MapContainer):
            def __init__(self, **kwds):
                super().__init__(**kwds)
                self.declare('a')
                self.declare('b', value=2)
                self.declare('c', value=3)

        d = container.MapContainer()
        d.declare('f')
        d.declare('g')
        d.declare('h')
        d.declare('i', value=container.ListContainer(container.UndefinedData))
        d.declare('j', value=container.ListContainer(LocalContainer), active=False)
        self.assertEqual(list(d.keys()), ['F', 'G', 'H', 'I', 'J'])

        self.assertEqual(
            """
I: 
""",
            str(d),
        )

        # Assigning to a field activates it
        d.f = 1
        self.assertEqual(
            """
F: 1
I: 
""",
            str(d),
        )
        self.assertEqual(d.f, 1)

        # Adding to a list also activates it, even if it was declared
        # with active=False
        self.assertTrue(d.i._active)
        self.assertFalse(d.j._active)
        d.j.add()
        self.assertTrue(d.i._active)
        self.assertTrue(d.j._active)
        self.assertEqual(
            """
F: 1
I: 
J: 
- B: 2
  C: 3
""",
            str(d),
        )

        self.assertEqual(
            """
- B: 2
  C: 3
""",
            str(d.j),
        )

    def test_pickle(self):
        d = container.MapContainer()
        d.declare('a', value=container.ignore)
        d.declare('b', value=container.undefined)
        d.declare('c', value=42)
        e = pickle.loads(pickle.dumps(d))
        self.assertEqual(list(d.keys()), list(e.keys()))
        self.assertIs(d.a, e.a)
        self.assertIs(d.b, e.b)
        self.assertEqual(d.c, e.c)

    def test_eq(self):
        d = container.MapContainer()
        d.declare('x', value=1)
        d.declare('y', value='a')
        d.declare('z', value=container.ListContainer(container.MapContainer))
        self.assertFalse(d == container.MapContainer())
        self.assertFalse(d == "Something else")

        e = container.ListContainer(container.UndefinedData)
        self.assertFalse(d == e)

        dd = container.MapContainer()
        dd.declare('x', value=1)
        dd.declare('y', value=1)
        dd.declare('z', value=container.ListContainer(container.UndefinedData))
        self.assertFalse(d == dd)
        d.y = 'b'
        dd.y = 'b'
        dd.declare('z', value=container.ListContainer(container.MapContainer))
        self.assertTrue(d == dd)

        d.z.add()
        d.z[0].declare('a', value=0)
        self.assertFalse(d == dd)
        dd.z.add()
        dd.z[0].declare('a', value=0)
        self.assertTrue(d == dd)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base import Transformation, TransformationFactory
import pyomo.gdp.plugins.hull


@TransformationFactory.register(
    'contrib.piecewise.multiple_choice',
    doc="Convert piecewise-linear model to a GDP "
    "to 'Multiple Choice' MIP "
    "formulation.",
)
class MultipleChoiceTransformation(Transformation):
    """
    Converts a model containing PiecewiseLinearFunctions to a an equivalent
    MIP via the Multiple Choice method from [1]. Note that,
    while this model probably resolves to the model described in [1] after
    presolve, the Pyomo version is not as simplified. Specifically, in [1], the
    the 'z' variables (representing the value of the piecewise-linear function
    in each Disjunct) are not disaggregated. In this transformation's output
    they will be, but a linear combination of inequalities yields a model
    equivalent to the Multiple Choice model in [1].

    References
    ----------
    [1] J.P. Vielma, S. Ahmed, and G. Nemhauser, "Mixed-integer models
        for nonseparable piecewise-linear optimization: unifying framework
        and extensions," Operations Research, vol. 58, no. 2, pp. 305-315,
        2010.
    """

    def _apply_to(self, instance, **kwds):
        TransformationFactory('contrib.piecewise.outer_repn_gdp').apply_to(instance)
        TransformationFactory('gdp.hull').apply_to(instance)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path
import subprocess

from pyomo.common.errors import ApplicationError
import pyomo.common
from pyomo.common.tempfiles import TempfileManager
from pyomo.opt.base import ProblemFormat, ConverterError
from pyomo.opt.base.convert import ProblemConverterFactory


@ProblemConverterFactory.register('ampl')
class AmplMIPConverter:
    def can_convert(self, from_type, to_type):
        """Returns true if this object supports the specified conversion"""
        #
        # Test if the ampl executable is available
        #
        if not pyomo.common.Executable("ampl"):
            return False
        #
        # Return True for specific from/to pairs
        #
        if from_type == ProblemFormat.mod and to_type == ProblemFormat.nl:
            return True
        if from_type == ProblemFormat.mod and to_type == ProblemFormat.mps:
            return True
        return False

    def apply(self, *args, **kwargs):
        """Convert an instance of one type into another"""
        if not isinstance(args[2], str):
            raise ConverterError("Can only apply ampl to convert file data")
        _exec = pyomo.common.Executable("ampl")
        if not _exec:
            raise ConverterError("The 'ampl' executable cannot be found")
        script_filename = TempfileManager.create_tempfile(suffix='.ampl')

        if args[1] == ProblemFormat.nl:
            output_filename = TempfileManager.create_tempfile(suffix='.nl')
        else:
            output_filename = TempfileManager.create_tempfile(suffix='.mps')

        cmd = [_exec.path(), script_filename]
        #
        # Create the AMPL script
        #
        OUTPUT = open(script_filename, 'w')
        OUTPUT.write("#\n")
        OUTPUT.write("# AMPL script for converting the following files\n")
        OUTPUT.write("#\n")
        if len(args[2:]) == 1:
            OUTPUT.write('model ' + args[2] + ";\n")
        else:
            OUTPUT.write('model ' + args[2] + ";\n")
            OUTPUT.write('data ' + args[3] + ";\n")
        abs_ofile = os.path.abspath(output_filename)
        if args[1] == ProblemFormat.nl:
            OUTPUT.write('write g' + abs_ofile[:-3] + ";\n")
        else:
            OUTPUT.write('write m' + abs_ofile[:-4] + ";\n")
        OUTPUT.close()
        #
        # Execute command and cleanup
        #
        output = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        if not os.path.exists(output_filename):  # pragma:nocover
            raise ApplicationError(
                "Problem launching 'ampl' to create '%s': %s"
                % (output_filename, output.stdout)
            )
        return (output_filename,), None  # empty variable map



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.opt.base import OptSolver
from pyomo.opt.base.solvers import SolverFactory
import logging

logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register('xpress', doc='The XPRESS LP/MIP solver')
class XPRESS(OptSolver):
    """The XPRESS LP/MIP solver"""

    def __new__(cls, *args, **kwds):
        mode = kwds.pop('solver_io', 'python')
        if mode is None:
            mode = 'python'

        if mode not in {'python', 'direct', 'persistent'}:
            logger.error(
                'Pyomo currently only supports a Python interface to XPRESS. '
                'Please use one of python, direct, or persistent for solver_io.'
            )
            return
        if mode in ['python', 'direct']:
            opt = SolverFactory('xpress_direct', **kwds)
        elif mode == 'persistent':
            opt = SolverFactory('xpress_persistent', **kwds)
        if opt is None:
            logger.error('Python API for XPRESS is not installed')
        return opt



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.opt.results.container import (
    ListContainer,
    MapContainer,
    UndefinedData,
    undefined,
    ignore,
)

from pyomo.opt.results.solver import (
    SolverStatus,
    TerminationCondition,
    check_optimal_termination,
    assert_optimal_termination,
)
from pyomo.opt.results.problem import ProblemSense
from pyomo.opt.results.solution import SolutionStatus, Solution
from pyomo.opt.results.results_ import SolverResults

from pyomo.common.deprecation import relocated_module_attribute

for _attr in ('ScalarData', 'ScalarType', 'default_print_options', 'strict'):
    relocated_module_attribute(
        _attr, 'pyomo.opt.results.container.' + _attr, version='6.8.1'
    )
del _attr
del relocated_module_attribute



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.config import (
    ConfigBlock,
    ConfigList,
    ConfigValue,
    In,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveInt,
)
from pyomo.common.deprecation import deprecation_warning
from pyomo.contrib.gdpopt.discrete_problem_initialize import valid_init_strategies
from pyomo.contrib.gdpopt.nlp_initialization import restore_vars_to_original_values
from pyomo.contrib.gdpopt.util import a_logger, _DoNothing
from pyomo.util.config_domains import ComponentDataSet
from pyomo.core.base import LogicalConstraint
from pyomo.gdp.disjunct import Disjunction

_supported_algorithms = {
    'LOA': ('gdpopt.loa', 'Logic-based Outer Approximation'),
    'GLOA': ('gdpopt.gloa', 'Global Logic-based Outer Approximation'),
    'LBB': ('gdpopt.lbb', 'Logic-based Branch and Bound'),
    'RIC': ('gdpopt.ric', 'Relaxation with Integer Cuts'),
    'enumerate': ('gdpopt.enumerate', 'Enumeration of discrete solutions'),
}


def _strategy_deprecation(strategy):
    deprecation_warning(
        "The argument 'strategy' has been deprecated in favor of 'algorithm.'",
        version="6.4.2",
    )
    return In(_supported_algorithms)(strategy)


def _init_strategy_deprecation(strategy):
    deprecation_warning(
        "The argument 'init_strategy' has been deprecated "
        "in favor of 'init_algorithm.'",
        version="6.4.2",
    )
    return In(valid_init_strategies)(strategy)


def _get_algorithm_config():
    CONFIG = ConfigBlock("GDPoptAlgorithm")
    CONFIG.declare(
        "strategy",
        ConfigValue(
            default=None,
            domain=_strategy_deprecation,
            description="DEPRECATED: Please use 'algorithm' instead.",
        ),
    )
    CONFIG.declare(
        "algorithm",
        ConfigValue(
            default=None,
            domain=In(_supported_algorithms),
            description="Algorithm to use.",
        ),
    )
    return CONFIG


def _add_common_configs(CONFIG):
    CONFIG.declare(
        "iterlim",
        ConfigValue(
            default=None, domain=NonNegativeInt, description="Iteration limit."
        ),
    )
    CONFIG.declare(
        "time_limit",
        ConfigValue(
            default=None,
            domain=PositiveInt,
            description="Time limit (seconds, default=600)",
            doc="""
            Seconds allowed until terminated. Note that the time limit can
            currently only be enforced between subsolver invocations. You may
            need to set subsolver time limits as well.""",
        ),
    )
    CONFIG.declare(
        "tee",
        ConfigValue(
            default=False, description="Stream output to terminal.", domain=bool
        ),
    )
    CONFIG.declare(
        "logger",
        ConfigValue(
            default='pyomo.contrib.gdpopt',
            description="The logger object or name to use for reporting.",
            domain=a_logger,
        ),
    )


def _add_nlp_solve_configs(CONFIG, default_nlp_init_method):
    # All of these config options are expected if the algorithm solves NLP
    # subproblems.
    CONFIG.declare(
        "integer_tolerance",
        ConfigValue(default=1e-5, description="Tolerance on integral values."),
    )
    CONFIG.declare(
        "constraint_tolerance",
        ConfigValue(
            default=1e-6,
            description="""
            Tolerance on constraint satisfaction.
    
            Increasing this tolerance corresponds to being more conservative in
            declaring the model or an NLP subproblem to be infeasible.
            """,
        ),
    )
    CONFIG.declare(
        "variable_tolerance",
        ConfigValue(default=1e-8, description="Tolerance on variable bounds."),
    )
    CONFIG.declare(
        "subproblem_initialization_method",
        ConfigValue(
            default=default_nlp_init_method,
            description=""""
            callback to specify custom routines to initialize the
            (MI)NLP subproblems.""",
            doc="""
            Callback to specify custom routines for initializing the (MI)NLP
            subproblems. This method is called after the discrete problem solution
            is fixed in the subproblem and before the subproblem is solved (or
            pre-solved).
    
            For algorithms with a discrete problem relaxation:
            This method accepts three arguments: the solver object, the subproblem
            GDPopt utility block and the discrete problem GDPopt utility block. The
            discrete problem contains the most recent discrete problem solution.
    
            For algorithms without a discrete problem relaxation:
            This method accepts four arguments: the list of Disjuncts that are
            currently fixed as being active, a list of values for the non-indicator
            BooleanVars (empty if force_nlp_subproblem=False), and a list of
            values for the integer vars (also empty if force_nlp_subproblem=False),
            and last the subproblem GDPopt utility block.
    
            The return of this method will be unused: The method should directly
            set the value of the variables on the subproblem
            """,
        ),
    )
    CONFIG.declare(
        "call_before_subproblem_solve",
        ConfigValue(
            default=_DoNothing,
            description="callback hook before calling the subproblem solver",
            doc="""
            Callback called right before the (MI)NLP subproblem is solved.
            Takes three arguments: The solver object, the subproblem and the
            GDPopt utility block on the subproblem.
    
            Note that unless you are *very* confident in what you are doing, the
            subproblem should not be modified in this callback: it should be used
            to interrogate the problem only.
    
            To initialize the problem before it is solved, please specify a method
            in the 'subproblem_initialization_method' argument.
            """,
        ),
    )
    CONFIG.declare(
        "call_after_subproblem_solve",
        ConfigValue(
            default=_DoNothing,
            description="""
            callback hook after a solution of the
            "nonlinear subproblem""",
            doc="""
            Callback called right after the (MI)NLP subproblem is solved.
            Takes three arguments: The solver object, the subproblem, and the
            GDPopt utility block on the subproblem.
    
            Note that unless you are *very* confident in what you are doing, the
            subproblem should not be modified in this callback: it should be used
            to interrogate the problem only.
            """,
        ),
    )
    CONFIG.declare(
        "call_after_subproblem_feasible",
        ConfigValue(
            default=_DoNothing,
            description="""
            callback hook after feasible solution of
            the nonlinear subproblem""",
            doc="""
            Callback called right after the (MI)NLP subproblem is solved,
            if it was feasible. Takes three arguments: The solver object, the
            subproblem and the GDPopt utility block on the subproblem.
    
            Note that unless you are *very* confident in what you are doing, the
            subproblem should not be modified in this callback: it should be used
            to interrogate the problem only.
            """,
        ),
    )
    CONFIG.declare(
        "force_subproblem_nlp",
        ConfigValue(
            default=False,
            description="""Force subproblems to be NLP, even if discrete variables
            exist.""",
        ),
    )
    CONFIG.declare(
        "subproblem_presolve",
        ConfigValue(
            default=True,
            description="""
        Flag to enable or disable subproblem presolve.
        Default=True.""",
            domain=bool,
        ),
    )
    CONFIG.declare(
        "tighten_nlp_var_bounds",
        ConfigValue(
            default=False,
            description="""
            Whether or not to do feasibility-based bounds tightening
            on the variables in the NLP subproblem before solving it.""",
            domain=bool,
        ),
    )
    CONFIG.declare(
        "round_discrete_vars",
        ConfigValue(
            default=True,
            description="""Flag to round subproblem discrete variable values to the
            nearest integer. Rounding is done before fixing disjuncts.""",
        ),
    )
    CONFIG.declare(
        "max_fbbt_iterations",
        ConfigValue(
            default=3,
            description="""
            Maximum number of feasibility-based bounds tightening
            iterations to do during NLP subproblem preprocessing.""",
            domain=PositiveInt,
        ),
    )


def _add_oa_configs(CONFIG):
    _add_nlp_solve_configs(
        CONFIG, default_nlp_init_method=restore_vars_to_original_values
    )

    CONFIG.declare(
        "init_strategy",
        ConfigValue(
            default=None,
            domain=_init_strategy_deprecation,
            description="DEPRECATED: Please use 'init_algorithm' instead.",
        ),
    )
    CONFIG.declare(
        "init_algorithm",
        ConfigValue(
            default="set_covering",
            domain=In(valid_init_strategies),
            description="Initialization algorithm to use.",
            doc="""
            Selects the initialization algorithm to use when generating
            the initial cuts to construct the discrete problem.""",
        ),
    )
    CONFIG.declare(
        "custom_init_disjuncts",
        ConfigList(
            # domain=ComponentSets of Disjuncts,
            default=None,
            description="List of disjunct sets to use for initialization.",
        ),
    )
    CONFIG.declare(
        "max_slack",
        ConfigValue(
            default=1000,
            domain=NonNegativeFloat,
            description="Upper bound on slack variables for OA",
        ),
    )
    CONFIG.declare(
        "OA_penalty_factor",
        ConfigValue(
            default=1000,
            domain=NonNegativeFloat,
            description="""
            Penalty multiplication term for slack variables on the
            objective value.""",
        ),
    )
    CONFIG.declare(
        "set_cover_iterlim",
        ConfigValue(
            default=8,
            domain=NonNegativeInt,
            description="Limit on the number of set covering iterations.",
        ),
    )
    CONFIG.declare(
        "discrete_problem_transformation",
        ConfigValue(
            default='gdp.bigm',
            description="""
            Name of the transformation to use to transform the
            discrete problem from a GDP to an algebraic model.""",
        ),
    )
    CONFIG.declare(
        "call_before_discrete_problem_solve",
        ConfigValue(
            default=_DoNothing,
            description="callback hook before calling the discrete problem solver",
            doc="""
            Callback called right before the MILP discrete problem is solved.
            Takes three arguments: The solver object, the discrete problem, and the
            GDPopt utility block on the discrete problem.
    
            Note that unless you are *very* confident in what you are doing, the
            problem should not be modified in this callback: it should be used
            to interrogate the problem only.
            """,
        ),
    )
    CONFIG.declare(
        "call_after_discrete_problem_solve",
        ConfigValue(
            default=_DoNothing,
            description="callback hook after a solution of the discrete problem",
            doc="""
            Callback called right after the MILP discrete problem is solved.
            Takes three arguments: The solver object, the discrete problem, and the
            GDPopt utility block on the discrete problem.
    
            Note that unless you are *very* confident in what you are doing, the
            problem should not be modified in this callback: it should be used
            to interrogate the problem only.
            """,
        ),
    )
    CONFIG.declare(
        "call_before_master_solve",
        ConfigValue(
            default=_DoNothing,
            description="DEPRECATED: Please use "
            "'call_before_discrete_problem_solve'",
        ),
    )
    CONFIG.declare(
        "call_after_master_solve",
        ConfigValue(
            default=_DoNothing,
            description="DEPRECATED: Please use 'call_after_discrete_problem_solve'",
        ),
    )
    CONFIG.declare(
        "mip_presolve",
        ConfigValue(
            default=True,
            description="""
        Flag to enable or disable GDPopt MIP presolve.
        Default=True.""",
            domain=bool,
        ),
    )
    CONFIG.declare(
        "calc_disjunctive_bounds",
        ConfigValue(
            default=False,
            description="""
        Calculate special disjunctive variable bounds for GLOA.
        False by default.""",
            domain=bool,
        ),
    )
    CONFIG.declare(
        "obbt_disjunctive_bounds",
        ConfigValue(
            default=False,
            description="""
            Use optimality-based bounds tightening rather than feasibility-based
            bounds tightening to compute disjunctive variable bounds. False by
            default.""",
            domain=bool,
        ),
    )


def _add_BB_configs(CONFIG):
    CONFIG.declare(
        "check_sat",
        ConfigValue(
            default=False,
            domain=bool,
            description="""
            When True, GDPopt-LBB will check satisfiability
            at each node via the pyomo.contrib.satsolver interface""",
        ),
    )
    CONFIG.declare(
        "solve_local_rnGDP",
        ConfigValue(
            default=False,
            domain=bool,
            description="""
            When True, GDPopt-LBB will solve a local MINLP at each node.""",
        ),
    )


def _add_mip_solver_configs(CONFIG):
    CONFIG.declare(
        "mip_solver",
        ConfigValue(
            default="gurobi",
            description="""
            Mixed-integer linear solver to use. Note that no persistent solvers
            other than the auto-persistent solvers in the APPSI package are
            supported.""",
        ),
    )
    CONFIG.declare(
        "mip_solver_args",
        ConfigBlock(
            description="""
            Keyword arguments to send to the MILP subsolver solve() invocation""",
            implicit=True,
        ),
    )


def _add_nlp_solver_configs(CONFIG, default_solver):
    CONFIG.declare(
        "nlp_solver",
        ConfigValue(
            default=default_solver,
            description="""
            Nonlinear solver to use. Note that no persistent solvers
            other than the auto-persistent solvers in the APPSI package are
            supported.""",
        ),
    )
    CONFIG.declare(
        "nlp_solver_args",
        ConfigBlock(
            description="""
            Keyword arguments to send to the NLP subsolver solve() invocation""",
            implicit=True,
        ),
    )
    CONFIG.declare(
        "minlp_solver",
        ConfigValue(
            default="baron",
            description="""
            Mixed-integer nonlinear solver to use. Note that no persistent solvers
            other than the auto-persistent solvers in the APPSI package are
            supported.""",
        ),
    )
    CONFIG.declare(
        "minlp_solver_args",
        ConfigBlock(
            description="""
            Keyword arguments to send to the MINLP subsolver solve() invocation""",
            implicit=True,
        ),
    )
    CONFIG.declare(
        "local_minlp_solver",
        ConfigValue(
            default="bonmin",
            description="""
            Mixed-integer nonlinear solver to use. Note that no persistent solvers
            other than the auto-persistent solvers in the APPSI package are
            supported.""",
        ),
    )
    CONFIG.declare(
        "local_minlp_solver_args",
        ConfigBlock(
            description="""
            Keyword arguments to send to the local MINLP subsolver solve()
            invocation""",
            implicit=True,
        ),
    )
    CONFIG.declare(
        "small_dual_tolerance",
        ConfigValue(
            default=1e-8,
            description="""
            When generating cuts, small duals multiplied by expressions can
            cause problems. Exclude all duals smaller in absolute value than the
            following.""",
        ),
    )


def _add_tolerance_configs(CONFIG):
    CONFIG.declare(
        "bound_tolerance",
        ConfigValue(
            default=1e-6,
            domain=NonNegativeFloat,
            description="Tolerance for bound convergence.",
        ),
    )


def _add_ldsda_configs(CONFIG):
    CONFIG.declare(
        "direction_norm",
        ConfigValue(
            default='L2',
            domain=In(['L2', 'Linf']),
            description="The norm to use for the search direction",
        ),
    )
    CONFIG.declare(
        "starting_point",
        ConfigValue(default=None, description="The value list of external variables."),
    )
    CONFIG.declare(
        "logical_constraint_list",
        ConfigValue(
            default=None,
            domain=ComponentDataSet(LogicalConstraint),
            description="""
            The list of logical constraints to be reformulated into external variables.
            The logical constraints should be in the same order of provided starting point.
            The provided logical constraints should be ExactlyExpressions.""",
        ),
    )
    CONFIG.declare(
        "disjunction_list",
        ConfigValue(
            default=None,
            domain=ComponentDataSet(Disjunction),
            description="""
            The list of disjunctions to be reformulated into external variables.
            The disjunctions should be in the same order of provided starting point.
            """,
        ),
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.environ import AbstractModel, Set, Param, Var, Objective, minimize

model = AbstractModel()

model.indices = Set(initialize=[1, 2])

model.p = Param(model.indices)

model.x = Var(model.indices)


def objective_rule(M):
    return sum([M.p[i] * M.x[i] for i in model.indices])


model.objective = Objective(rule=objective_rule, sense=minimize)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from . import (
    numvalue,
    visitor,
    numeric_expr,
    boolean_value,
    logical_expr,
    relational_expr,
)

#
# FIXME: remove circular dependencies between logical_expr and numeric_expr
#

# Initialize logicalvalue functions
boolean_value._generate_logical_proposition = logical_expr._generate_logical_proposition


from pyomo.common.numeric_types import (
    value,
    native_numeric_types,
    native_types,
    nonpyomo_leaf_types,
)
from pyomo.common.errors import TemplateExpressionError

from .base import ExpressionBase
from .boolean_value import BooleanValue
from .expr_common import ExpressionType, Mode, OperatorAssociativity
from .logical_expr import (
    native_logical_types,
    special_boolean_atom_types,
    #
    BooleanValue,
    BooleanConstant,
    BooleanExpression,
    BooleanExpressionBase,
    #
    UnaryBooleanExpression,
    NotExpression,
    BinaryBooleanExpression,
    EquivalenceExpression,
    XorExpression,
    ImplicationExpression,
    NaryBooleanExpression,
    AndExpression,
    OrExpression,
    ExactlyExpression,
    AtMostExpression,
    AtLeastExpression,
    AllDifferentExpression,
    CountIfExpression,
    #
    land,
    lnot,
    lor,
    xor,
    equivalent,
    exactly,
    atleast,
    atmost,
    all_different,
    count_if,
    implies,
)
from .numeric_expr import (
    NumericValue,
    NumericExpression,
    # operators:
    AbsExpression,
    DivisionExpression,
    Expr_ifExpression,
    ExternalFunctionExpression,
    LinearExpression,
    MaxExpression,
    MinExpression,
    MonomialTermExpression,
    NegationExpression,
    PowExpression,
    ProductExpression,
    SumExpressionBase,  # TODO: deprecate / remove
    SumExpression,
    UnaryFunctionExpression,
    # TBD: remove export of NPV classes here?
    NPV_AbsExpression,
    NPV_DivisionExpression,
    NPV_Expr_ifExpression,
    NPV_ExternalFunctionExpression,
    NPV_MaxExpression,
    NPV_MinExpression,
    NPV_NegationExpression,
    NPV_PowExpression,
    NPV_ProductExpression,
    NPV_SumExpression,
    NPV_UnaryFunctionExpression,
    # functions to generate expressions
    Expr_if,
    log,
    log10,
    sin,
    cos,
    tan,
    cosh,
    sinh,
    tanh,
    asin,
    acos,
    atan,
    exp,
    sqrt,
    asinh,
    acosh,
    atanh,
    ceil,
    floor,
    # Lgacy utilities
    NPV_expression_types,  # TODO: remove
    LinearDecompositionError,  # TODO: move to common.errors
    decompose_term,
    linear_expression,
    nonlinear_expression,
    mutable_expression,
)
from .numvalue import (
    as_numeric,
    is_constant,
    is_fixed,
    is_variable_type,
    is_potentially_variable,
    ZeroConstant,
    polynomial_degree,
)
from .relational_expr import (
    RelationalExpression,
    RangedExpression,
    InequalityExpression,
    EqualityExpression,
    NotEqualExpression,
    inequality,
)
from .symbol_map import SymbolMap
from .template_expr import (
    GetItemExpression,
    Numeric_GetItemExpression,
    Boolean_GetItemExpression,
    Structural_GetItemExpression,
    GetAttrExpression,
    Numeric_GetAttrExpression,
    Boolean_GetAttrExpression,
    Structural_GetAttrExpression,
    CallExpression,
    TemplateSumExpression,
    #
    NPV_Numeric_GetItemExpression,
    NPV_Boolean_GetItemExpression,
    NPV_Structural_GetItemExpression,
    NPV_Numeric_GetAttrExpression,
    NPV_Boolean_GetAttrExpression,
    NPV_Structural_GetAttrExpression,
    #
    IndexTemplate,
    resolve_template,
    ReplaceTemplateExpression,
    substitute_template_expression,
    substitute_getitem_with_param,
    substitute_template_with_value,
    templatize_rule,
    templatize_constraint,
)
from .visitor import (
    StreamBasedExpressionVisitor,
    SimpleExpressionVisitor,
    ExpressionValueVisitor,
    ExpressionReplacementVisitor,
    FixedExpressionError,
    NonConstantExpressionError,
    identify_components,
    identify_variables,
    identify_mutable_parameters,
    clone_expression,
    evaluate_expression,
    expression_to_string,
    polynomial_degree,
    replace_expressions,
    sizeof_expression,
)

from .calculus.derivatives import differentiate
from .taylor_series import taylor_series_expansion

#
# declare deprecation paths for removed modules and attributes
#
from pyomo.common.deprecation import moved_module

moved_module(
    "pyomo.core.expr.current",
    "pyomo._archive.current",
    msg="pyomo.core.expr.current is deprecated.  "
    "Please import expression symbols from pyomo.core.expr",
    version='6.6.2',
)
del moved_module



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

import os
import time

import pyomo.common.unittest as unittest

import pyomo.dataportal.parse_datacmds as parser
from pyomo.common.errors import DeveloperError


class TestDatParser(unittest.TestCase):
    def test_arguments(self):
        self.assertIsNone(parser.parse_data_commands(data=None, filename=None))
        self.assertEqual(parser.parse_data_commands(data='', filename=None), {None: []})

    def test_update_parsetable(self):
        def _bad_sig():
            return "122345"

        _orig = parser._get_this_file_signature, parser.dat_lexer, parser.dat_yaccer
        try:
            parser._get_this_file_signature = _bad_sig
            parser.dat_lexer = None
            with self.assertRaisesRegex(
                DeveloperError,
                r"DAT parse tables \(pyomo.dataportal._parse_table_datacmds\) "
                r"out of\s+sync with parser definition",
            ):
                parser.parse_data_commands('')
        finally:
            parser._get_this_file_signature, parser.dat_lexer, parser.dat_yaccer = _orig



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from pyomo.common.collections import ComponentMap
from pyomo.core.base import Block, Var, Constraint, Objective, Suffix, value
from pyomo.core.plugins.transform.hierarchy import Transformation
from pyomo.core.base import TransformationFactory
from pyomo.core.base.suffix import SuffixFinder
from pyomo.core.expr import replace_expressions
from pyomo.util.components import rename_components

logger = logging.getLogger("pyomo.core.plugins.transform.scaling")


@TransformationFactory.register(
    'core.scale_model', doc="Scale model variables, constraints, and objectives."
)
class ScaleModel(Transformation):
    """
    Transformation to scale a model.

    This plugin performs variable, constraint, and objective scaling on
    a model based on the scaling factors in the suffix 'scaling_factor'
    set for the variables, constraints, and/or objective. This is typically
    done to scale the problem for improved numerical properties.

    Supported transformation methods:
        * :py:meth:`apply_to <pyomo.core.plugins.transform.scaling.ScaleModel.apply_to>`
        * :py:meth:`create_using <pyomo.core.plugins.transform.scaling.ScaleModel.create_using>`
        * :py:meth:`propagate_solution <pyomo.core.plugins.transform.scaling.ScaleModel.propagate_solution>`

    By default, scaling components are renamed with the prefix ``scaled_``. To disable
    this behavior and scale variables in-place (or keep the same names in a new model),
    use the ``rename=False`` argument to ``apply_to`` or ``create_using``.


    Examples
    --------

    .. doctest::

        >>> from pyomo.environ import *
        >>> # create the model
        >>> model = ConcreteModel()
        >>> model.x = Var(bounds=(-5, 5), initialize=1.0)
        >>> model.y = Var(bounds=(0, 1), initialize=1.0)
        >>> model.obj = Objective(expr=1e8*model.x + 1e6*model.y)
        >>> model.con = Constraint(expr=model.x + model.y == 1.0)
        >>> # create the scaling factors
        >>> model.scaling_factor = Suffix(direction=Suffix.EXPORT)
        >>> model.scaling_factor[model.obj] = 1e-6 # scale the objective
        >>> model.scaling_factor[model.con] = 2.0  # scale the constraint
        >>> model.scaling_factor[model.x] = 0.2    # scale the x variable
        >>> # transform the model
        >>> scaled_model = TransformationFactory('core.scale_model').create_using(model)
        >>> # print the value of the objective function to show scaling has occurred
        >>> print(value(model.x))
        1.0
        >>> print(value(scaled_model.scaled_x))
        0.2
        >>> print(value(scaled_model.scaled_x.lb))
        -1.0
        >>> print(value(model.obj))
        101000000.0
        >>> print(value(scaled_model.scaled_obj))
        101.0

    """

    def __init__(self, **kwds):
        kwds['name'] = "scale_model"
        self._scaling_method = kwds.pop('scaling_method', 'user')
        self._suffix_finder = None
        super(ScaleModel, self).__init__(**kwds)

    def _create_using(self, original_model, **kwds):
        scaled_model = original_model.clone()
        self._apply_to(scaled_model, **kwds)
        return scaled_model

    def _apply_to(self, model, rename=True):
        # create a map of component to scaling factor
        component_scaling_factor_map = ComponentMap()
        self._suffix_finder = SuffixFinder('scaling_factor', 1.0, model)

        # if the scaling_method is 'user', get the scaling parameters from the suffixes
        if self._scaling_method == 'user':
            # get the scaling factors
            for c in model.component_data_objects(
                ctype=(Var, Constraint, Objective), descend_into=True
            ):
                component_scaling_factor_map[c] = self._suffix_finder.find(c)
        else:
            raise ValueError(
                "ScaleModel transformation: unknown scaling_method found"
                "-- supported values: 'user' "
            )

        if rename:
            # rename all the Vars, Constraints, and Objectives
            # from foo to scaled_foo
            component_list = list(
                model.component_objects(ctype=[Var, Constraint, Objective])
            )
            scaled_component_to_original_name_map = rename_components(
                model=model, component_list=component_list, prefix='scaled_'
            )
        else:
            scaled_component_to_original_name_map = ComponentMap(
                [
                    (comp, comp.name)
                    for comp in model.component_objects(
                        ctype=[Var, Constraint, Objective]
                    )
                ]
            )

        # scale the variable bounds and values and build the variable
        # substitution map for scaling vars in constraints
        variable_substitution_map = ComponentMap()
        already_scaled = set()
        for variable in [
            var for var in model.component_objects(ctype=Var, descend_into=True)
        ]:
            if variable.is_reference():
                # Skip any references - these should get picked up when
                # handling the actual variable
                continue

            # set the bounds/value for the scaled variable
            for k in variable:
                v = variable[k]
                if id(v) in already_scaled:
                    continue
                already_scaled.add(id(v))
                scaling_factor = component_scaling_factor_map[v]
                variable_substitution_map[v] = v / scaling_factor

                if v.lb is not None:
                    v.setlb(v.lb * scaling_factor)
                if v.ub is not None:
                    v.setub(v.ub * scaling_factor)
                if scaling_factor < 0:
                    temp = v.lb
                    v.setlb(v.ub)
                    v.setub(temp)

                if v.value is not None:
                    # Since the value was OK in the unscaled space, it
                    # should be safe to assume it is still valid in the
                    # scaled space)
                    v.set_value(value(v) * scaling_factor, skip_validation=True)

        # scale the objectives/constraints and perform the scaled variable substitution
        scale_constraint_dual = False
        if type(model.component('dual')) is Suffix:
            scale_constraint_dual = True

        # translate the variable_substitution_map (ComponentMap)
        # to variable_substitution_dict (key: id() of component)
        # ToDo: We should change replace_expressions to accept a ComponentMap as well
        variable_substitution_dict = {
            id(k): v for k, v in variable_substitution_map.items()
        }

        already_scaled = set()
        for component in model.component_objects(
            ctype=(Constraint, Objective), descend_into=True
        ):
            if component.is_reference():
                # Skip any references - these should get picked up when
                # handling the actual component
                continue

            for k in component:
                c = component[k]
                if id(c) in already_scaled:
                    continue
                already_scaled.add(id(c))
                # perform the constraint/objective scaling and variable sub
                scaling_factor = component_scaling_factor_map[c]
                if c.ctype is Constraint:
                    body = scaling_factor * replace_expressions(
                        expr=c.body,
                        substitution_map=variable_substitution_dict,
                        descend_into_named_expressions=True,
                        remove_named_expressions=True,
                    )

                    # scale the rhs
                    lower = c.lower
                    upper = c.upper
                    if lower is not None:
                        lower = lower * scaling_factor
                    if upper is not None:
                        upper = upper * scaling_factor

                    if scaling_factor < 0:
                        lower, upper = upper, lower

                    if scale_constraint_dual and c in model.dual:
                        dual_value = model.dual[c]
                        if dual_value is not None:
                            model.dual[c] = dual_value / scaling_factor

                    if c.equality:
                        c.set_value((lower, body))
                    else:
                        c.set_value((lower, body, upper))

                elif c.ctype is Objective:
                    c.expr = scaling_factor * replace_expressions(
                        expr=c.expr,
                        substitution_map=variable_substitution_dict,
                        descend_into_named_expressions=True,
                        remove_named_expressions=True,
                    )
                else:
                    raise NotImplementedError(
                        'Unknown object type found when applying scaling factors '
                        'in ScaleModel transformation - Internal Error'
                    )

        model.component_scaling_factor_map = component_scaling_factor_map
        model.scaled_component_to_original_name_map = (
            scaled_component_to_original_name_map
        )

        # Now that we have scaled the model, deactivate the relevant
        # scaling suffixes so that we don't accidentally (later)
        # double-scale.
        for s in self._suffix_finder.all_suffixes:
            s.deactivate()

        return model

    def propagate_solution(self, scaled_model, original_model):
        """This method takes the solution in scaled_model and maps it back to
        the original model.

        It will also transform duals and reduced costs if the suffixes
        'dual' and/or 'rc' are present.  The :code:`scaled_model`
        argument must be a model that was already scaled using this
        transformation as it expects data from the transformation to
        perform the back mapping.

        Parameters
        ----------
        scaled_model : Pyomo Model
           The model that was previously scaled with this transformation
        original_model : Pyomo Model
           The original unscaled source model

        """
        if not hasattr(scaled_model, 'component_scaling_factor_map'):
            raise AttributeError(
                'ScaleModel:propagate_solution called with scaled_model that does '
                'not have a component_scaling_factor_map. It is possible this '
                'method was called using a model that was not scaled with the '
                'ScaleModel transformation'
            )
        if not hasattr(scaled_model, 'scaled_component_to_original_name_map'):
            raise AttributeError(
                'ScaleModel:propagate_solution called with scaled_model that does '
                'not have a scaled_component_to_original_name_map. It is possible '
                'this method was called using a model that was not scaled with '
                'the ScaleModel transformation'
            )

        component_scaling_factor_map = scaled_model.component_scaling_factor_map
        scaled_component_to_original_name_map = (
            scaled_model.scaled_component_to_original_name_map
        )

        # transfer the variable values and reduced costs
        check_reduced_costs = type(scaled_model.component('rc')) is Suffix
        check_dual = (
            type(scaled_model.component('dual')) is Suffix
            and type(original_model.component('dual')) is Suffix
        )

        if check_reduced_costs or check_dual:
            # get the objective scaling factor
            scaled_objectives = list(
                scaled_model.component_data_objects(
                    ctype=Objective, active=True, descend_into=True
                )
            )
            if len(scaled_objectives) != 1:
                raise NotImplementedError(
                    'ScaleModel.propagate_solution requires a single active '
                    'objective function, but %d objectives found.'
                    % (len(scaled_objectives))
                )
            else:
                objective_scaling_factor = component_scaling_factor_map[
                    scaled_objectives[0]
                ]

        for scaled_v in scaled_model.component_objects(ctype=Var, descend_into=True):
            # get the unscaled_v from the original model
            original_v_path = scaled_component_to_original_name_map[scaled_v]
            # This will not work if decimal indices are present:
            original_v = original_model.find_component(original_v_path)

            for k in scaled_v:
                if scaled_v[k].value is None and original_v[k].value is not None:
                    logger.warning(
                        "Variable with value None in the scaled model is replacing"
                        f" value of variable {original_v[k].name} in the original"
                        f" model with None (was {original_v[k].value})."
                    )
                    original_v[k].set_value(None, skip_validation=True)
                elif scaled_v[k].value is not None:
                    original_v[k].set_value(
                        value(scaled_v[k]) / component_scaling_factor_map[scaled_v[k]],
                        skip_validation=True,
                    )
                if check_reduced_costs and scaled_v[k] in scaled_model.rc:
                    original_model.rc[original_v[k]] = (
                        scaled_model.rc[scaled_v[k]]
                        * component_scaling_factor_map[scaled_v[k]]
                        / objective_scaling_factor
                    )

        # transfer the duals
        if check_dual:
            for scaled_c in scaled_model.component_objects(
                ctype=Constraint, descend_into=True
            ):
                original_c = original_model.find_component(
                    scaled_component_to_original_name_map[scaled_c]
                )

                for k in scaled_c:
                    original_model.dual[original_c[k]] = (
                        scaled_model.dual[scaled_c[k]]
                        * component_scaling_factor_map[scaled_c[k]]
                        / objective_scaling_factor
                    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pytest
import random

import pyomo.environ as pyo
import pyomo.dae as dae
from pyomo.common.collections import ComponentMap
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.series_data import TimeSeriesData
from pyomo.contrib.mpc.data.interval_data import IntervalData
from pyomo.contrib.mpc.data.convert import (
    _process_to_dynamic_data,
    interval_to_series,
    series_to_interval,
)


def _make_model():
    m = pyo.ConcreteModel()
    m.time = pyo.Set(initialize=[0.1 * i for i in range(11)])
    m.comp = pyo.Set(initialize=["A", "B"])
    m.var = pyo.Var(m.time, m.comp, initialize=1.0)
    return m


class TestIntervalToSeries(unittest.TestCase):
    def test_no_time_points(self):
        m = _make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.7, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)

        series_data = interval_to_series(interval_data)
        # Default uses right endpoint of each interval
        pred_time_points = [0.2, 0.5, 1.0]
        pred_data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        self.assertEqual(series_data, TimeSeriesData(pred_data, pred_time_points))

    def test_no_time_points_left_endpoints(self):
        m = _make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.7, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)

        series_data = interval_to_series(interval_data, use_left_endpoints=True)
        pred_time_points = [0.0, 0.2, 0.7]
        pred_data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        self.assertEqual(series_data, TimeSeriesData(pred_data, pred_time_points))

    def test_time_points_provided_no_boundary(self):
        m = _make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)

        # Choose some time points that don't lie on interval boundaries
        time_points = [0.05 + i * 0.1 for i in range(10)]
        series_data = interval_to_series(interval_data, time_points=time_points)
        pred_data = {
            m.var[:, "A"]: [1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0],
            m.var[:, "B"]: [4.0, 4.0, 5.0, 5.0, 5.0, 6.0, 6.0, 6.0, 6.0, 6.0],
        }
        self.assertEqual(series_data, TimeSeriesData(pred_data, time_points))

    def test_time_points_provided_some_on_boundary(self):
        m = _make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)

        time_points = [0.1 * i for i in range(11)]
        series_data = interval_to_series(interval_data, time_points=time_points)
        # Some of the time points are on interval boundaries. By default we
        # use the values from the intervals on the left.
        pred_data = {
            m.var[:, "A"]: [1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0],
            m.var[:, "B"]: [4.0, 4.0, 4.0, 5.0, 5.0, 5.0, 6.0, 6.0, 6.0, 6.0, 6.0],
        }
        self.assertEqual(series_data, TimeSeriesData(pred_data, time_points))

    def test_time_points_provided_some_on_boundary_use_right(self):
        m = _make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)

        time_points = [0.1 * i for i in range(11)]
        series_data = interval_to_series(
            interval_data, time_points=time_points, prefer_left=False
        )
        # Some of the time points are on interval boundaries. By default we
        # use the values from the intervals on the left.
        pred_data = {
            m.var[:, "A"]: [1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0],
            m.var[:, "B"]: [4.0, 4.0, 5.0, 5.0, 5.0, 6.0, 6.0, 6.0, 6.0, 6.0, 6.0],
        }
        self.assertEqual(series_data, TimeSeriesData(pred_data, time_points))

    def test_with_roundoff_error(self):
        m = _make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)

        # Simulate roundoff error in these time points.
        random.seed(12710)
        time_points = [i * 0.1 + random.uniform(-1e-8, 1e-8) for i in range(11)]
        series_data = interval_to_series(
            interval_data, time_points=time_points, tolerance=1e-7
        )
        pred_data = {
            m.var[:, "A"]: [1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0],
            m.var[:, "B"]: [4.0, 4.0, 4.0, 5.0, 5.0, 5.0, 6.0, 6.0, 6.0, 6.0, 6.0],
        }
        self.assertEqual(series_data, TimeSeriesData(pred_data, time_points))


class TestSeriesToInterval(unittest.TestCase):
    def test_singleton(self):
        m = _make_model()
        time_points = [0.1]
        data = {m.var[:, "A"]: [0.5], m.var[:, "B"]: [2.0]}
        series_data = TimeSeriesData(data, time_points)
        interval_data = series_to_interval(series_data)
        pred_data = IntervalData(
            {m.var[:, "A"]: [0.5], m.var[:, "B"]: [2.0]}, [(0.1, 0.1)]
        )
        self.assertEqual(interval_data, pred_data)

    def test_convert(self):
        m = _make_model()
        time_points = [0.1, 0.2, 0.3, 0.4, 0.5]
        data = {
            m.var[:, "A"]: [1.0, 2.0, 3.0, 4.0, 5.0],
            m.var[:, "B"]: [6.0, 7.0, 8.0, 9.0, 10.0],
        }
        series_data = TimeSeriesData(data, time_points)
        interval_data = series_to_interval(series_data)

        pred_data = IntervalData(
            {m.var[:, "A"]: [2.0, 3.0, 4.0, 5.0], m.var[:, "B"]: [7.0, 8.0, 9.0, 10.0]},
            [(0.1, 0.2), (0.2, 0.3), (0.3, 0.4), (0.4, 0.5)],
        )
        self.assertEqual(pred_data, interval_data)

    def test_convert_use_right(self):
        m = _make_model()
        time_points = [0.1, 0.2, 0.3, 0.4, 0.5]
        data = {
            m.var[:, "A"]: [1.0, 2.0, 3.0, 4.0, 5.0],
            m.var[:, "B"]: [6.0, 7.0, 8.0, 9.0, 10.0],
        }
        series_data = TimeSeriesData(data, time_points)
        interval_data = series_to_interval(series_data, use_left_endpoints=True)

        pred_data = IntervalData(
            {m.var[:, "A"]: [1.0, 2.0, 3.0, 4.0], m.var[:, "B"]: [6.0, 7.0, 8.0, 9.0]},
            [(0.1, 0.2), (0.2, 0.3), (0.3, 0.4), (0.4, 0.5)],
        )
        self.assertEqual(pred_data, interval_data)


class TestProcessToDynamic(unittest.TestCase):
    def test_non_time_indexed_data(self):
        m = _make_model()
        m.scalar_var = pyo.Var(m.comp, initialize=3.0)
        data = ComponentMap([(m.scalar_var["A"], 3.1), (m.scalar_var["B"], 3.2)])
        # Passing non-time-indexed data to ScalarData just returns
        # a ScalarData object with the non-time-indexed CUIDs as keys.
        dyn_data = _process_to_dynamic_data(data)
        self.assertTrue(isinstance(dyn_data, ScalarData))
        self.assertIn(pyo.ComponentUID(m.scalar_var["A"]), dyn_data.get_data())
        self.assertIn(pyo.ComponentUID(m.scalar_var["B"]), dyn_data.get_data())


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import random

import pyomo.environ as pyo


def multi_knapsack(
    num_items: int = 20, item_ub: int = 10, num_cons: int = 10, max_weight: int = 100
) -> pyo.ConcreteModel:
    "Creates a random instance of Knapsack with multiple capacity constraints."
    mod = pyo.ConcreteModel()
    mod.I = pyo.Set(initialize=range(num_items), name="I")
    mod.J = pyo.Set(initialize=range(num_cons), name="I")

    rng = random.Random(0)
    weight = [[rng.randint(0, max_weight) for _ in mod.I] for _ in mod.J]
    cost = [rng.random() for _ in mod.I]
    capacity = 0.1 * num_items * item_ub * max_weight

    mod.x = pyo.Var(mod.I, domain=pyo.Integers, bounds=(0, item_ub), name="x")
    mod.cap = pyo.Constraint(
        mod.J,
        rule=lambda m, j: sum(weight[j][i] * m.x[i] for i in m.I) <= capacity,
        name="cap",
    )
    mod.obj = pyo.Objective(
        rule=lambda m: sum(cost[i] * m.x[i] for i in m.I), sense=pyo.maximize
    )
    return mod



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import operator

from pyomo.common.collections import ComponentMap
from pyomo.common.config import In
from pyomo.common.deprecation import deprecated
from pyomo.common.enums import IntEnum
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.common.pyomo_typing import overload
from pyomo.common.timing import ConstructionTimer
from pyomo.core.base.block import BlockData
from pyomo.core.base.component import ActiveComponent, ModelComponentFactory
from pyomo.core.base.disable_methods import disable_methods
from pyomo.core.base.enums import SortComponents
from pyomo.core.base.initializer import Initializer

logger = logging.getLogger('pyomo.core')

_SUFFIX_API = (
    ('__contains__', 'test membership in'),
    ('__iter__', 'iterate over'),
    '__getitem__',
    '__setitem__',
    'set_value',
    'set_all_values',
    'clear_value',
    'clear_all_values',
    'update_values',
)

# A list of convenient suffix generators, including:
#   - active_export_suffix_generator
#       **(used by problem writers)
#   - export_suffix_generator
#   - active_import_suffix_generator
#       **(used by OptSolver and PyomoModel._load_solution)
#   - import_suffix_generator
#   - active_local_suffix_generator
#   - local_suffix_generator
#   - active_suffix_generator
#   - suffix_generator


def suffix_generator(a_block, datatype=NOTSET, direction=NOTSET, active=None):
    _iter = (
        (s.local_name, s)
        for s in a_block.component_data_objects(
            Suffix, active=active, descend_into=False
        )
    )
    if direction is not NOTSET:
        direction = _SuffixDirectionDomain(direction)
        if not direction:
            _iter = filter(lambda item: item[1].direction == direction, _iter)
        else:
            _iter = filter(lambda item: item[1].direction & direction, _iter)
    if datatype is not NOTSET:
        _iter = filter(lambda item: item[1].datatype == datatype, _iter)
    return _iter


def active_export_suffix_generator(a_block, datatype=NOTSET):
    return suffix_generator(a_block, datatype, SuffixDirection.EXPORT, True)


def export_suffix_generator(a_block, datatype=NOTSET):
    return suffix_generator(a_block, datatype, SuffixDirection.EXPORT)


def active_import_suffix_generator(a_block, datatype=NOTSET):
    return suffix_generator(a_block, datatype, SuffixDirection.IMPORT, True)


def import_suffix_generator(a_block, datatype=NOTSET):
    return suffix_generator(a_block, datatype, SuffixDirection.IMPORT)


def active_local_suffix_generator(a_block, datatype=NOTSET):
    return suffix_generator(a_block, datatype, SuffixDirection.LOCAL, True)


def local_suffix_generator(a_block, datatype=NOTSET):
    return suffix_generator(a_block, datatype, SuffixDirection.LOCAL)


def active_suffix_generator(a_block, datatype=NOTSET):
    return suffix_generator(a_block, datatype, active=True)


class SuffixDataType(IntEnum):
    """Suffix data types

    AMPL only supports two data types for Suffixes: int and float.  The
    numeric values here are specific to the NL file format and should
    not be changed without checking/updating the NL writer.

    """

    INT = 0
    FLOAT = 4


class SuffixDirection(IntEnum):
    """Suffix data flow definition.

    This identifies if the specific Suffix is to be sent to the solver,
    read from the solver output, both, or neither:

    - LOCAL: Suffix is local to Pyomo and should not be sent to or read
      from the solver.

    - EXPORT: Suffix should be sent to the solver as supplemental model
      information.

    - IMPORT: Suffix values will be returned from the solver and should
      be read from the solver output.

    - IMPORT_EXPORT: The Suffix is both an EXPORT and IMPORT suffix.

    """

    LOCAL = 0
    EXPORT = 1
    IMPORT = 2
    IMPORT_EXPORT = 3


_SuffixDataTypeDomain = In(SuffixDataType)
_SuffixDirectionDomain = In(SuffixDirection)


@ModelComponentFactory.register("Declare a container for extraneous model data")
class Suffix(ComponentMap, ActiveComponent):
    """A model suffix, representing extraneous model data"""

    """
    Constructor Arguments:
        direction   The direction of information flow for this suffix.
                        By default, this is LOCAL, indicating that no
                        suffix data is exported or imported.
        datatype    A variable type associated with all values of this
                        suffix.
    """

    #
    # The following local (class) aliases are provided for convenience
    # and backwards compatibility with The Book, 3rd ed
    #

    # Suffix Directions:
    # - neither sent to solver or received from solver
    LOCAL = SuffixDirection.LOCAL
    # - sent to solver or other external location
    EXPORT = SuffixDirection.EXPORT
    # - obtained from solver or other external source
    IMPORT = SuffixDirection.IMPORT
    # - both import and export
    IMPORT_EXPORT = SuffixDirection.IMPORT_EXPORT

    FLOAT = SuffixDataType.FLOAT
    INT = SuffixDataType.INT

    def __new__(cls, *args, **kwargs):
        if cls is not Suffix:
            return super().__new__(cls)
        return super().__new__(AbstractSuffix)

    def __setstate__(self, state):
        super().__setstate__(state)
        # As the concrete class *is* the "Suffix" base class, the normal
        # implementation of deepcopy (through get/setstate) will create
        # the new Suffix, and __new__ will map it to AbstractSuffix.  We
        # need to map constructed Suffixes back to Suffix:
        if self._constructed and self.__class__ is AbstractSuffix:
            self.__class__ = Suffix

    @overload
    def __init__(
        self,
        *,
        direction=LOCAL,
        datatype=FLOAT,
        initialize=None,
        rule=None,
        name=None,
        doc=None,
    ): ...

    def __init__(self, **kwargs):
        # Suffix type information
        self._direction = None
        self._datatype = None
        self._rule = None

        # The suffix direction (note the setter performs error checking)
        self.direction = kwargs.pop('direction', Suffix.LOCAL)

        # The suffix datatype (note the setter performs error checking)
        self.datatype = kwargs.pop('datatype', Suffix.FLOAT)

        # The suffix construction rule
        # TODO: deprecate the use of 'rule'
        self._rule = Initializer(
            self._pop_from_kwargs('Suffix', kwargs, ('rule', 'initialize'), None),
            treat_sequences_as_mappings=False,
            allow_generators=True,
        )

        # Initialize base classes
        kwargs.setdefault('ctype', Suffix)
        ActiveComponent.__init__(self, **kwargs)
        ComponentMap.__init__(self)

        if self._rule is None:
            self.construct()

    def construct(self, data=None):
        """
        Constructs this component, applying rule if it exists.
        """
        if is_debug_set(logger):
            logger.debug(f"Constructing %s '%s'", self.__class__.__name__, self.name)

        if self._constructed is True:
            return

        timer = ConstructionTimer(self)
        self._constructed = True

        if self._rule is not None:
            rule = self._rule
            if rule.contains_indices():
                # The rule contains explicit indices (e.g., is a dict).
                # Iterate over the indices, expand them, and store the
                # result
                block = self.parent_block()
                for index in rule.indices():
                    self.set_value(index, rule(block, index), expand=True)
            else:
                self.update_values(rule(self.parent_block(), None), expand=True)
        timer.report()

    @property
    def datatype(self):
        """Return the suffix datatype."""
        return self._datatype

    @datatype.setter
    def datatype(self, datatype):
        """Set the suffix datatype."""
        if datatype is not None:
            datatype = _SuffixDataTypeDomain(datatype)
        self._datatype = datatype

    @property
    def direction(self):
        """Return the suffix direction."""
        return self._direction

    @direction.setter
    def direction(self, direction):
        """Set the suffix direction."""
        self._direction = _SuffixDirectionDomain(direction)

    def export_enabled(self):
        """
        Returns True when this suffix is enabled for export to
        solvers.
        """
        return bool(self._direction & Suffix.EXPORT)

    def import_enabled(self):
        """
        Returns True when this suffix is enabled for import from
        solutions.
        """
        return bool(self._direction & Suffix.IMPORT)

    def update_values(self, data, expand=True):
        """
        Updates the suffix data given a list of component,value
        tuples. Provides an improvement in efficiency over calling
        set_value on every component.
        """
        if expand:
            try:
                items = data.items()
            except AttributeError:
                items = data

            for component, value in items:
                self.set_value(component, value, expand=expand)

        else:
            # As implemented by MutableMapping
            self.update(data)

    def set_value(self, component, value, expand=True):
        """
        Sets the value of this suffix on the specified component.

        When expand is True (default), array components are handled by
        storing a reference and value for each index, with no
        reference being stored for the array component itself. When
        expand is False (this is the case for __setitem__), this
        behavior is disabled and a reference to the array component
        itself is kept.
        """
        if expand and component.is_indexed():
            for component_ in component.values():
                self[component_] = value
        else:
            self[component] = value

    def set_all_values(self, value):
        """
        Sets the value of this suffix on all components.
        """
        for ndx in self:
            self[ndx] = value

    def clear_value(self, component, expand=True):
        """
        Clears suffix information for a component.
        """
        if expand and component.is_indexed():
            for component_ in component.values():
                self.pop(component_, None)
        else:
            self.pop(component, None)

    def clear_all_values(self):
        """
        Clears all suffix data.
        """
        self.clear()

    @deprecated(
        'Suffix.set_datatype is replaced with the Suffix.datatype property',
        version='6.7.1',
    )
    def set_datatype(self, datatype):
        """
        Set the suffix datatype.
        """
        self.datatype = datatype

    @deprecated(
        'Suffix.get_datatype is replaced with the Suffix.datatype property',
        version='6.7.1',
    )
    def get_datatype(self):
        """
        Return the suffix datatype.
        """
        return self.datatype

    @deprecated(
        'Suffix.set_direction is replaced with the Suffix.direction property',
        version='6.7.1',
    )
    def set_direction(self, direction):
        """
        Set the suffix direction.
        """
        self.direction = direction

    @deprecated(
        'Suffix.get_direction is replaced with the Suffix.direction property',
        version='6.7.1',
    )
    def get_direction(self):
        """
        Return the suffix direction.
        """
        return self.direction

    def _pprint(self):
        def _data(sort):
            data = ((str(k), v) for k, v in self._dict.values())
            if SortComponents.SORTED_INDICES in sort:
                data = sorted(data, key=operator.itemgetter(0))
            return data

        return (
            [
                ('Direction', str(self._direction.name)),
                ('Datatype', getattr(self._datatype, 'name', 'None')),
            ],
            _data,
            ("Value",),
            lambda k, v: [v],
        )

    #
    # Override a few methods to make sure the ActiveComponent versions are
    # called. We can't just switch the inheritance order due to
    # complications with __setstate__
    #

    def __str__(self):
        return ActiveComponent.__str__(self)


@disable_methods(_SUFFIX_API)
class AbstractSuffix(Suffix):
    pass


class SuffixFinder:
    def __init__(self, name, default=None, context=None):
        """This provides an efficient utility for finding suffix values on a
        (hierarchical) Pyomo model.

        Parameters
        ----------
        name: str

            Name of :py:class:`Suffix` to search for.

        default:

            Default value to return from `.find()` if no matching Suffix
            is found.

        context: BlockData

            The root of the Block hierarchy to use when searching for
            Suffix components.  Suffixes outside this hierarchy will not
            be interrogated and components that are queried (with
            :py:meth:`find(component_data)` will return the default
            value.

        """
        self.name = name
        self.default = default
        self.all_suffixes = []
        self._context = context
        self._suffixes_by_block = ComponentMap()
        self._suffixes_by_block[self._context] = []
        if context is not None:
            s = context.component(name)
            if s is not None and s.ctype is Suffix and s.active:
                self._suffixes_by_block[context].append(s)
                self.all_suffixes.append(s)

    def find(self, component_data):
        """Find suffix value for a given component data object in model tree

        If the `component_data` is attached to a block within the
        `context` (that is, part of the model hierarchy rooted by the
        `context` :py:class:`BlockData`), then :py:class:`Suffix` components are
        searched by traversing the model hierarchy in three passes:

        1. Search for a :py:class:`Suffix` matching the specific `component_data`,
           starting at the `root` (context block) and descending down
           the tree to the `component_data`.  Return the first match
           found.

        2. Search for a :py:class:`Suffix` matching the `component_data`'s
           container, starting at the `root` and descending down the
           tree to the `component_data`.  Return the first match found.

        3. Search for a :py:class:`Suffix` with key ``None``, starting from the
           `component_data` and working up the tree to the `root`.
           Return the first match found.

        If the `component_data` is not part of the `context` hierarchy,
        then only the :py:class:`Suffix` attached to the root context is
        checked for the `component_data`, its container, and ``None`` (in
        order).

        If no valid value is found by searching :py:class:`Suffix` components,
        then the `default` value is returned.

        Parameters
        ----------
        component_data: ComponentBase

            Component or component data object to find suffix value for.

        Returns
        -------
        Any
            The value for Suffix associated with component data if found, else None.

        """
        # Walk parent tree and search for suffixes
        if isinstance(component_data, BlockData):
            _block = component_data
        else:
            _block = component_data.parent_block()
        try:
            suffixes = self._get_suffix_list(_block)
        except AttributeError:
            # Component was outside the context (eventually parent
            # becomes None and parent.parent_block() raises an
            # AttributeError).
            #
            # We will check the context's suffix to see if there is an
            # entry there.  We will look in the same order: component
            # data, component, and None
            #
            # TBD: Should we check for / accept None?  The simple
            # solution (and probably correct answer) is yes.
            suffixes = self._suffixes_by_block[self._context]

        # Pass 1: look for the component_data, working root to leaf
        for s in suffixes:
            if component_data in s:
                return s[component_data]
        # Pass 2: look for the component container, working root to leaf
        parent_comp = component_data.parent_component()
        if parent_comp is not component_data:
            for s in suffixes:
                if parent_comp in s:
                    return s[parent_comp]
        # Pass 3: look for None, working leaf to root
        for s in reversed(suffixes):
            if None in s:
                return s[None]
        return self.default

    def _get_suffix_list(self, parent):
        if parent in self._suffixes_by_block:
            return self._suffixes_by_block[parent]
        suffixes = list(self._get_suffix_list(parent.parent_block()))
        self._suffixes_by_block[parent] = suffixes
        s = parent.component(self.name)
        if s is not None and s.ctype is Suffix and s.active:
            suffixes.append(s)
            self.all_suffixes.append(s)
        return suffixes



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import namedtuple
from pyomo.common.dependencies import networkx as nx
from pyomo.contrib.incidence_analysis.common.dulmage_mendelsohn import (
    dulmage_mendelsohn as dm_nx,
)

"""
This module imports the general Dulmage-Mendelsohn-on-a-graph function
from "common" and implements an interface for coo_matrix-like objects.
"""

RowPartition = namedtuple(
    "RowPartition", ["unmatched", "overconstrained", "underconstrained", "square"]
)
"""Named tuple containing the subsets of the Dulmage-Mendelsohn partition
when applied to matrix rows (constraints).

"""

ColPartition = namedtuple(
    "ColPartition", ["unmatched", "underconstrained", "overconstrained", "square"]
)
"""Named tuple containing the subsets of the Dulmage-Mendelsohn partition
when applied to matrix columns (variables).

"""


def dulmage_mendelsohn(matrix_or_graph, top_nodes=None, matching=None):
    """Partition a bipartite graph or incidence matrix according to the
    Dulmage-Mendelsohn characterization

    The Dulmage-Mendelsohn partition tells which nodes of the two bipartite
    sets *can possibly be* unmatched after a maximum cardinality matching.
    Applied to an incidence matrix, it can be interpreted as partitioning
    rows and columns into under-constrained, over-constrained, and
    well-constrained subsystems.

    As it is often useful to explicitly check the unmatched rows and columns,
    ``dulmage_mendelsohn`` partitions rows into the subsets:

    - **underconstrained** - The rows matched with *possibly* unmatched
      columns (unmatched and underconstrained columns)
    - **square** - The well-constrained rows, which are matched with
      well-constrained columns
    - **overconstrained** - The matched rows that *can possibly be* unmatched
      in some maximum cardinality matching
    - **unmatched** - The unmatched rows in a particular maximum cardinality
      matching

    and partitions columns into the subsets:

    - **unmatched** - The unmatched columns in a particular maximum cardinality
      matching
    - **underconstrained** - The columns that *can possibly be* unmatched in
      some maximum cardinality matching
    - **square** - The well-constrained columns, which are matched with
      well-constrained rows
    - **overconstrained** - The columns matched with *possibly* unmatched
      rows (unmatched and overconstrained rows)

    While the Dulmage-Mendelsohn decomposition does not specify an order within
    any of these subsets, the order returned by this function preserves the
    maximum matching that is used to compute the decomposition. That is, zipping
    "corresponding" row and column subsets yields pairs in this maximum matching.
    For example:

    .. doctest::
       :hide:
       :skipif: not (networkx_available and scipy_available)

       >>> # Hidden code block to make the following example runnable
       >>> import scipy.sparse as sps
       >>> from pyomo.contrib.incidence_analysis.dulmage_mendelsohn import dulmage_mendelsohn
       >>> matrix = sps.identity(3)

    .. doctest::
       :skipif: not (networkx_available and scipy_available)

       >>> row_dmpartition, col_dmpartition = dulmage_mendelsohn(matrix)
       >>> rdmp = row_dmpartition
       >>> cdmp = col_dmpartition
       >>> matching = list(zip(
       ...     rdmp.underconstrained + rdmp.square + rdmp.overconstrained,
       ...     cdmp.underconstrained + cdmp.square + cdmp.overconstrained,
       ... ))
       >>> # matching is a valid maximum matching of rows and columns of the matrix!

    Parameters
    ----------
    matrix_or_graph: ``scipy.sparse.coo_matrix`` or ``networkx.Graph``
        The incidence matrix or bipartite graph to be partitioned
    top_nodes: list
        List of nodes in one bipartite set of the graph. Must be provided
        if a graph is provided.
    matching: dict
        A maximum cardinality matching in the form of a dict mapping
        from "top nodes" to their matched nodes *and* from the matched
        nodes back to the "top nodes".

    Returns
    -------
    row_dmp: RowPartition
        The Dulmage-Mendelsohn partition of rows
    col_dmp: ColPartition
        The Dulmage-Mendelsohn partition of columns

    """
    if isinstance(matrix_or_graph, nx.Graph):
        # The purpose of handling graphs here is that if we construct NX graphs
        # directly from Pyomo expressions, we can eliminate the overhead of
        # converting expressions to a matrix, then the matrix to a graph.
        #
        # In this case, top_nodes should correspond to constraints.
        graph = matrix_or_graph
        if top_nodes is None:
            raise ValueError(
                "top_nodes must be specified if a graph is provided,"
                "\notherwise the result is ambiguous."
            )
        partition = dm_nx(graph, top_nodes=top_nodes, matching=matching)
        # RowPartition and ColPartition do not make sense for a general graph.
        # However, here we assume that this graph comes from a Pyomo model,
        # and that "top nodes" are constraints.
        partition = (RowPartition(*partition[0]), ColPartition(*partition[1]))
    else:
        # Assume matrix_or_graph is a scipy coo_matrix
        matrix = matrix_or_graph
        M, N = matrix.shape
        nxb = nx.algorithms.bipartite
        from_biadjacency_matrix = nxb.matrix.from_biadjacency_matrix

        if matching is not None:
            # If a matching was provided for a COO matrix, we assume it
            # maps row indices to column indices, for compatibility with
            # output of our maximum_matching function.

            # NetworkX graph has column nodes offset by M
            matching = {i: j + M for i, j in matching.items()}
            inv_matching = {j: i for i, j in matching.items()}
            # DM function requires matching map to contain inverse matching too
            matching.update(inv_matching)

        # Matrix rows have bipartite=0, columns have bipartite=1
        bg = from_biadjacency_matrix(matrix)
        row_partition, col_partition = dm_nx(
            bg, top_nodes=list(range(M)), matching=matching
        )

        partition = (
            row_partition,
            tuple([n - M for n in subset] for subset in col_partition),
            # Column nodes have values in [M, M+N-1]. Apply the offset
            # to get values corresponding to indices in user's matrix.
        )

    partition = (RowPartition(*partition[0]), ColPartition(*partition[1]))
    return partition



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.contrib.pynumero.interfaces.external_grey_box import ExternalGreyBoxBlock
from pyomo.contrib.pynumero.examples.external_grey_box.react_example.reactor_model_residuals import (
    ReactorModel,
    ReactorModelWithHessian,
    ReactorModelNoOutputs,
    ReactorModelScaled,
    create_pyomo_reactor_model,
)


def maximize_cb_ratio_residuals_with_output(
    show_solver_log=False, additional_options={}
):
    # in this simple example, we will use an external grey box model representing
    # a steady-state reactor, and solve for the space velocity that maximizes
    # the ratio of B to the other components coming out of the reactor
    # This example illustrates the use of "equality constraints" or residuals
    # in the external grey box example as well as outputs
    m = pyo.ConcreteModel()

    # create a block to store the external reactor model
    m.reactor = ExternalGreyBoxBlock(external_model=ReactorModel())

    # The feed concentration will be fixed for this example
    m.cafcon = pyo.Constraint(expr=m.reactor.inputs['caf'] == 10000)

    # add an objective function that maximizes the concentration
    # of cb coming out of the reactor
    m.obj = pyo.Objective(expr=m.reactor.outputs['cb_ratio'], sense=pyo.maximize)

    solver = pyo.SolverFactory('cyipopt')
    solver.config.options['hessian_approximation'] = 'limited-memory'
    for k, v in additional_options.items():
        solver.config.options[k] = v
    results = solver.solve(m, tee=show_solver_log)
    pyo.assert_optimal_termination(results)
    return m


def maximize_cb_ratio_residuals_with_hessian_with_output(
    show_solver_log=False, additional_options={}
):
    # in this simple example, we will use an external grey box model representing
    # a steady-state reactor, and solve for the space velocity that maximizes
    # the ratio of B to the other components coming out of the reactor
    # This example illustrates the use of "equality constraints" or residuals
    # in the external grey box example as well as outputs
    m = pyo.ConcreteModel()

    # create a block to store the external reactor model
    m.reactor = ExternalGreyBoxBlock(external_model=ReactorModelWithHessian())

    # The feed concentration will be fixed for this example
    m.cafcon = pyo.Constraint(expr=m.reactor.inputs['caf'] == 10000)

    # add an objective function that maximizes the concentration
    # of cb coming out of the reactor
    m.obj = pyo.Objective(expr=m.reactor.outputs['cb_ratio'], sense=pyo.maximize)

    solver = pyo.SolverFactory('cyipopt')
    for k, v in additional_options.items():
        solver.config.options[k] = v
    results = solver.solve(m, tee=show_solver_log)
    pyo.assert_optimal_termination(results)
    return m


def maximize_cb_ratio_residuals_with_hessian_with_output_pyomo(
    show_solver_log=False, additional_options={}
):
    # this example is the same as the one above, but solves with a pure
    # pyomo model - this is mostly for comparison and testing
    m = create_pyomo_reactor_model()
    solver = pyo.SolverFactory('ipopt')
    for k, v in additional_options.items():
        solver.options[k] = v
    solver.options['linear_solver'] = 'mumps'
    results = solver.solve(m, tee=show_solver_log)
    pyo.assert_optimal_termination(results)
    return m


def maximize_cb_ratio_residuals_with_output_scaling(
    show_solver_log=False, additional_options={}
):
    # in this simple example, we will use an external grey box model representing
    # a steady-state reactor, and solve for the space velocity that maximizes
    # the ratio of B to the other components coming out of the reactor
    # This example illustrates the use of "equality constraints" or residuals
    # in the external grey box example as well as outputs

    # This example also shows how to do scaling.
    # There are two things to scale, the "Pyomo" variables/constraints,
    # and the external model residuals / output equations.
    # The scaling factors for the external residuals and output equations
    # are set in the derived ExternalGreyBoxModel class (see ReactorModelScaled
    # for this example).
    # The scaling factors for the Pyomo part of the model are set using suffixes
    # - this requires that we declare the scaling suffix on the main model as
    #   shown below.
    # - Then the scaling factors can be set directly on the Pyomo variables
    #   and constraints as also shown below.
    # - Note: In this example, the scaling factors for the input and output
    #   variables from the grey box model are set in the finalize_block_construction
    #   callback (again, see ReactorModelScaled)
    m = pyo.ConcreteModel()

    # declare the scaling suffix on the model
    m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

    # create a block to store the external reactor model
    m.reactor = ExternalGreyBoxBlock(external_model=ReactorModelScaled())

    # The feed concentration will be fixed for this example
    m.cafcon = pyo.Constraint(expr=m.reactor.inputs['caf'] == 10000)
    # set a scaling factor for this constraint - if we had additional pyomo
    # variables, we could set them the same way
    m.scaling_factor[m.cafcon] = 42.0

    # add an objective function that maximizes the concentration
    # of cb coming out of the reactor
    m.obj = pyo.Objective(expr=m.reactor.outputs['cb_ratio'], sense=pyo.maximize)

    solver = pyo.SolverFactory('cyipopt')
    solver.config.options['hessian_approximation'] = 'limited-memory'
    for k, v in additional_options.items():
        solver.config.options[k] = v
    results = solver.solve(m, tee=show_solver_log)
    pyo.assert_optimal_termination(results)
    return m


def maximize_cb_ratio_residuals_with_obj(show_solver_log=False, additional_options={}):
    # in this simple example, we will use an external grey box model representing
    # a steady-state reactor, and solve for the space velocity that maximizes
    # the ratio of B to the other components coming out of the reactor
    # This example illustrates the use of "equality constraints" or residuals
    # in the external grey box example as well as additional pyomo variables
    # and constraints
    m = pyo.ConcreteModel()

    # create a block to store the external reactor model
    m.reactor = ExternalGreyBoxBlock()
    m.reactor.set_external_model(ReactorModelNoOutputs())

    # The feed concentration will be fixed for this example
    m.cafcon = pyo.Constraint(expr=m.reactor.inputs['caf'] == 10000)

    # add an objective function that maximizes the concentration
    # of cb coming out of the reactor
    u = m.reactor.inputs
    m.obj = pyo.Objective(
        expr=u['cb'] / (u['ca'] + u['cc'] + u['cd']), sense=pyo.maximize
    )

    solver = pyo.SolverFactory('cyipopt')
    solver.config.options['hessian_approximation'] = 'limited-memory'
    for k, v in additional_options.items():
        solver.config.options[k] = v
    results = solver.solve(m, tee=show_solver_log)
    pyo.assert_optimal_termination(results)
    return m


def maximize_cb_ratio_residuals_with_pyomo_variables(
    show_solver_log=False, additional_options={}
):
    # in this simple example, we will use an external grey box model representing
    # a steady-state reactor, and solve for the space velocity that maximizes
    # the ratio of B to the other components coming out of the reactor
    # This example illustrates the use of "equality constraints" or residuals
    # in the external grey box example as well as additional pyomo variables
    # and constraints
    m = pyo.ConcreteModel()

    # create a block to store the external reactor model
    m.reactor = ExternalGreyBoxBlock()
    m.reactor.set_external_model(ReactorModelNoOutputs())

    # add a variable and constraint for the cb ratio
    m.cb_ratio = pyo.Var(initialize=1)
    u = m.reactor.inputs
    m.cb_ratio_con = pyo.Constraint(
        expr=u['cb'] / (u['ca'] + u['cc'] + u['cd']) - m.cb_ratio == 0
    )

    # The feed concentration will be fixed for this example
    m.cafcon = pyo.Constraint(expr=m.reactor.inputs['caf'] == 10000)

    # add an objective function that maximizes the concentration
    # of cb coming out of the reactor
    m.obj = pyo.Objective(expr=m.cb_ratio, sense=pyo.maximize)

    solver = pyo.SolverFactory('cyipopt')
    solver.config.options['hessian_approximation'] = 'limited-memory'
    for k, v in additional_options.items():
        solver.config.options[k] = v
    results = solver.solve(m, tee=show_solver_log)
    pyo.assert_optimal_termination(results)
    return m


if __name__ == '__main__':
    m = maximize_cb_ratio_residuals_with_output(show_solver_log=True)

    # the next two are the same model with pyomo/ipopt and external/cyipopt
    m = maximize_cb_ratio_residuals_with_hessian_with_output_pyomo(show_solver_log=True)
    m = maximize_cb_ratio_residuals_with_hessian_with_output(show_solver_log=True)

    aoptions = {
        'hessian_approximation': 'limited-memory',
        #'limited_memory_update_type': 'sr1',
        'nlp_scaling_method': 'user-scaling',
        'print_level': 10,
    }
    m = maximize_cb_ratio_residuals_with_output_scaling(
        show_solver_log=True, additional_options=aoptions
    )

    m = maximize_cb_ratio_residuals_with_obj(show_solver_log=True)

    m = maximize_cb_ratio_residuals_with_pyomo_variables(show_solver_log=True)
